
// This demo program solves Poisson's equation,
//
//		- div grad u(x, y) = f(x, y)
//
// on the unit square with source f given by
//
//		f(x, y) = 500*exp(-((x-0.5)^2 + (y-0.5)^2)/0.02)
//
// and boundary conditions given by
//
//		u(x, y)		= 0
//		du/dn(x, y)	= 0
//
// using a discontinuous Galerkin formulation (interior penalty method).

#include <dolfin.h>
#include <fmg.h>
#include "Poisson.h"

using namespace dolfin;

// Source term
class Source : public Expression
{
public:
	Source() : Expression() {}

	void eval(Array<double>& values, const Array<double>& x) const
	{
		const double dx = x[0] - 0.5;
		const double dy = x[1] - 0.5;
		values[0] = 500.0*exp(-(dx*dx + dy*dy)/0.02);
	}

};

int main(int argc, char* argv[])
{
	// Create mesh and function space
	UnitSquareMesh mesh(8, 8);
	Poisson::FunctionSpace V(mesh);

	// Create functions
	Source f;

	// Define variational problem
	Poisson::BilinearForm a(V, V);
	Poisson::LinearForm L(V);
	L.f = f;

	Function u(V);
	LinearVariationalProblem problem(a, L, u);
	LinearVariationalSolver solver(problem);

	fmg::Tests tests(problem);
	tests.parameters["num_refinements"] = 4;
	tests.parameters.parse(argc, argv);
	tests.run();

	return 0;
}
