
// This demo tests the prolongation operator for various element types,
// by assembling a bilinear form on two levels (A_c and A_f) and comparing the matrix on the coarse level with
// the coarse matrix computed by the Galerkin approximation A = P^T A_f P, using the maximum norm ||A-A_c||_linf.

#include <dolfin.h>
#include <fmg.h>
#include "Tests.h"

using namespace dolfin;


// Compute ||A-A_c||_linf.
void test(const Form& form0)
{
	const FunctionSpace& V0 = *(form0.function_space(0));
	const Mesh& mesh0 = *(V0.mesh());
	const Mesh& mesh1 = adapt(mesh0);
	const Form& form1 = adapt(form0, reference_to_no_delete_pointer(mesh1));
	const FunctionSpace& V1 = *(form1.function_space(0));

	boost::shared_ptr<GenericMatrix> A1Gen(PETScFactory::instance().create_matrix());
	PETScMatrix& A1 = *dynamic_cast<PETScMatrix*>(A1Gen.get());
	dolfin::assemble(A1, form1);

	boost::shared_ptr<dolfin::GenericMatrix> PGen = fmg::ProlongationAssembler::assemble_matrix(V0, V1, PETScFactory::instance());
	const PETScMatrix& P = *dynamic_cast<PETScMatrix*>(PGen.get());

	Mat* pA0 = new Mat();
	MatPtAP(*(A1.mat()), *(P.mat()), MAT_INITIAL_MATRIX, 1.0, pA0);
	PETScMatrix A0(boost::shared_ptr<Mat>(pA0, dolfin::PETScMatrixDeleter()));

	// this code prints information if Galerkin assembly method creates the same lhs as
	// assemble_system by printing the difference in linf norm

	boost::shared_ptr<GenericMatrix> dAGen(PETScFactory::instance().create_matrix());
	PETScMatrix& dA = *(dynamic_cast<PETScMatrix*>(dAGen.get()));
	dolfin::assemble(dA, form0);
	dA -= A0;

	std::cout << V0.element()->signature() << ": " << dA.norm("linf") << std::endl;
}


#define RUN_TEST(name, mesh) do { \
		name::FunctionSpace V(mesh); \
		test(name::BilinearForm(V, V)); } while (false)


int main()
{
	std::size_t n = 2;
	UnitIntervalMesh mesh1(n);
	UnitCircleMesh mesh2(n);
	UnitSphereMesh mesh3(n);

	SubSystemsManager::init_petsc();

	#include "RunTests.h"

	return 0;
}
