
// This code compares a geometric Multigrid with Hypre AMG for a Poisson problem.

#include <dolfin.h>
#include "Poisson.h"
#include <fmg.h>

using namespace dolfin;

// hierarchical multigrid class
class mg_level : public Hierarchical<mg_level>
{
public:
	boost::shared_ptr<const FunctionSpace> space;		// function space
	boost::shared_ptr<GenericMatrix> P;			// prolongation operator from this level to child
	boost::shared_ptr<GenericMatrix> A;			// system matrix
	boost::shared_ptr<GenericVector> b;			// system rhs
	boost::shared_ptr<GenericVector> x;			// solution
	boost::shared_ptr<GenericVector> r;			// residual
	boost::shared_ptr<fmg::GenericSmoothingOperator> S1;	// smoother
	boost::shared_ptr<fmg::GenericSmoothingOperator> S2;	// smoother
	boost::shared_ptr<GenericLinearSolver> Ai;		// coarse level solver
	DirichletBC::Map dbc_map;				// Dirichlet boundary condition (global dof to value)
	bool x_assumed_zero;					// flag indicating if x should be zero

	mg_level() : Hierarchical<mg_level>(*this)
	{
		A.reset(new PETScMatrix());
		b.reset(new PETScVector());
		x.reset(new PETScVector());
		r.reset(new PETScVector());
		x_assumed_zero = false;
	}

	// perform recursive V-cycle multigrid
	// same pre- and post-smoother with one smooting iteration
	void recursive_mg()
	{
		Timer t("recursive_mg");

		// get coarser level
		boost::shared_ptr<mg_level> p_level = parent_shared_ptr();

		if (!p_level) {
			// we are on coarsest level
			Ai->solve(*x, *b);
			return;
		}

		// apply pre-smoother
		S1->apply(*x, *b, x_assumed_zero, 1);

		// compute residual r = A*x - b
		if (x_assumed_zero)
			r->zero();
		else
			A->mult(*x, *r);
		(*r) -= (*b);

		// restrict residual to coarser level child.b = P^T*r
		zero_dbc(*r);
		p_level->P->transpmult(*r, *(p_level->b));
		p_level->zero_dbc(*(p_level->b));

		// do recursive multigrid
		p_level->x_assumed_zero = true;
		p_level->recursive_mg();

		// prolongate coarse level solution
		// (we just use the residual vector for storage of the correction)
		p_level->zero_dbc(*(p_level->x));
		p_level->P->mult(*(p_level->x), *r);
		zero_dbc(*r);

		// correct solution
		if (x_assumed_zero) {
			(*r) *= -1;
			(*x) = (*r);
		}
		else
			(*x) -= (*r);

		// apply post-smoother
		S2->apply(*x, *b, x_assumed_zero, 1);
	}

	void recursive_mg(GenericVector& x, const GenericVector& b, bool x_assumed_zero = true)
	{
		boost::shared_ptr<GenericVector> x_org(this->x);
		boost::shared_ptr<GenericVector> b_org(this->b);

		this->x_assumed_zero = x_assumed_zero;
		this->x = reference_to_no_delete_pointer(x);
		this->b = reference_to_no_delete_pointer(*(const_cast<GenericVector*>(&b)));

		recursive_mg();

		this->x = x_org;
		this->b = b_org;
	}

	void zero_dbc(GenericVector& x)
	{
		DirichletBC::Map::const_iterator i;

		for (i = dbc_map.begin(); i != dbc_map.end(); i++) {
			x.setitem(i->first, i->second);
		}
		x.apply("insert");
	}

	virtual void init_smoother()
	{
		dolfin_assert(A && A->size(0));

		S1.reset(new fmg::SmoothingOperator(A, "fsor", 1.0));
		S2.reset(new fmg::SmoothingOperator(A, "bsor", 1.0));
	}

	virtual void init_galerkin_A()
	{
		boost::shared_ptr<mg_level> child = child_shared_ptr();

		dolfin_assert(child);
		dolfin_assert(child->A && child->A->size(0));
		dolfin_assert(P && P->size(0));

		Mat* pA0 = new Mat();
		MatPtAP(*(dynamic_cast<dolfin::PETScMatrix*>(child->A.get()))->mat(),
				*(dynamic_cast<dolfin::PETScMatrix*>(P.get()))->mat(),
				MAT_INITIAL_MATRIX, 1.0, pA0);
		A.reset(new dolfin::PETScMatrix(boost::shared_ptr<Mat>(pA0, dolfin::PETScMatrixDeleter())));
	}

	virtual void init_prolongation()
	{
		boost::shared_ptr<mg_level> child = child_shared_ptr();

		dolfin_assert(child && child->space);

		P = fmg::ProlongationAssembler::assemble_matrix(
			*space, *(child->space), PETScFactory::instance(),
			fmg::ProlongationAssembler::AssemblyMethods::uniform);
	}

	virtual void init_solver()
	{
		Ai.reset(new PETScLUSolver());
		Ai->set_operator(A);
	}

	void set_child(boost::shared_ptr<mg_level> child)
	{
		Hierarchical<mg_level>::set_child(child);
		child->set_parent(reference_to_no_delete_pointer(*this));
	}
};


// multigrid preconditioner
class mg_prec : public PETScUserPreconditioner
{
public:
	mg_prec(boost::shared_ptr<mg_level> level0) : _level0(level0) { }
	
	void solve(PETScVector& x, const PETScVector& b)
	{
		// apply one multigrid iteration
		_level0->recursive_mg(x, b);
	}

protected:
	boost::shared_ptr<mg_level> _level0;
};


// Source term (right-hand side)
class Source : public Expression
{
	void eval(Array<double>& values, const Array<double>& x) const
	{
		double dx = x[0] - 0.5;
		double dy = x[1] - 0.5;
		values[0] = 10*exp(-(dx*dx + dy*dy) / 0.02);
	}
};

// Normal derivative (Neumann boundary condition)
class dUdN : public Expression
{
	void eval(Array<double>& values, const Array<double>& x) const
	{
		values[0] = sin(5*x[0]);
	}
};

// Sub domain for Dirichlet boundary condition
class DirichletBoundary : public SubDomain
{
	bool inside(const Array<double>& x, bool on_boundary) const
	{
		return x[0] < DOLFIN_EPS or x[0] > 1.0 - DOLFIN_EPS;
	}
};


boost::shared_ptr<LinearVariationalProblem> create_problem(std::size_t level)
{
	// Create mesh and function space
	boost::shared_ptr<const Mesh> mesh(new UnitSquareMesh(7 << level, 7 << level));
	boost::shared_ptr<FunctionSpace> V(new Poisson::FunctionSpace(mesh));

	// Define boundary condition
	boost::shared_ptr<Expression> u0(new Constant(0.0));
	boost::shared_ptr<SubDomain> boundary(new DirichletBoundary());
	boost::shared_ptr<DirichletBC> bc(new DirichletBC(V, u0, boundary, "pointwise"));

	// Define variational forms
	boost::shared_ptr<Poisson::BilinearForm> a(new Poisson::BilinearForm(V, V));
	boost::shared_ptr<Poisson::LinearForm> L(new Poisson::LinearForm(V));
	boost::shared_ptr<Expression> f(new Source());
	boost::shared_ptr<Expression> g(new dUdN());
	L->set_coefficient("f", f);
	L->set_coefficient("g", g);

	// Define variational problem
	boost::shared_ptr<Function> u(new Function(V));
	std::vector<boost::shared_ptr<const BoundaryCondition> > bcs;
	bcs.push_back(bc);
	boost::shared_ptr<LinearVariationalProblem> problem(new LinearVariationalProblem(a, L, u, bcs));

	return problem;
}



int main(int argc, char* argv[])
{
	std::size_t nrefine = 5;
	double rtol = 1e-6;

	// set some options for Hypre AMG
	// see http://www.gnu-darwin.org/www001/ports-1.5a-CURRENT/math/petsc/work/petsc-2.3.3-p0/src/ksp/pc/impls/hypre/hypre.c.html
	// for more possibilities
	//
	// CycleTypes   = {"","V","W"};
	// CoarsenTypes = {"CLJP","Ruge-Stueben","","modifiedRuge-Stueben","","","Falgout", "", "PMIS", "", "HMIS"};
	// MeasureTypes = {"local","global"};
	// RelaxTypes   = {"Jacobi","sequential-Gauss-Seidel","","SOR/Jacobi","backward-SOR/Jacobi","","symmetric-SOR/Jacobi","","","Gaussian-elimination"};
	// InterpTypes  = {"classical", "", "", "direct", "multipass", "multipass-wts", "ext+i", "ext+i-cc", "standard", "standard-wts", "", "", "FF", "FF1"};

	const char* params[] = {argv[0],
		//	"--petsc.pc_hypre_boomeramg_relax_type_up", "SOR/Jacobi",
		//	"--petsc.pc_hypre_boomeramg_relax_type_down", "backward-SOR/Jacobi",
		//	"--petsc.pc_hypre_boomeramg_max_iter", "1",
		//	"--petsc.pc_hypre_boomeramg_rtol", "0.0",
		//	"--petsc.pc_hypre_boomeramg_strong_threshold", "0.0",
		//	"--petsc.pc_hypre_boomeramg_coarsen_type", "CLJP",
		//	"--petsc.pc_hypre_boomeramg_max_levels", "10",
		//	"--petsc.pc_hypre_boomeramg_agg_nl", "1",
		//	"--petsc.pc_hypre_boomeramg_relax_type_all", "SOR",
		//	"--petsc.pc_hypre_boomeramg_print_statistics", "1",
		//	"--petsc.pc_hypre_boomeramg_interp_type", "standard",
	};
	parameters.parse(sizeof(params)/sizeof(char *), const_cast<char**>(params));

	// disable dof reordering
	parameters["reorder_dofs_serial"] = false;

	// solve problem with geometric multigrid
	{
		std::cout << "~~~~~~~~~~~~~~~~~ FMG ~~~~~~~~~~~~~~~~~" << std::endl;

		Timer tt("total multigrid");
		Timer ti("init multigrid");

		// init levels
		std::vector<boost::shared_ptr<mg_level> > levels;
		levels.resize(nrefine+1);

		boost::shared_ptr<LinearVariationalProblem> problem = create_problem(0);
		LinearVariationalProblem* problem_i = problem.get();
		const FunctionSpace* space_i = problem_i->test_space().get();
		const Mesh* mesh_i = space_i->mesh().get();

		for (std::size_t i = 0; i < levels.size(); i++)
		{
			// create new level
			levels[i].reset(new mg_level());

			if (i > 0)
			{
				// set parent/child relation between levels
				levels[i-1]->set_child(levels[i]);

				// adapt mesh
				fmg::Adaptor::adapt(*mesh_i);
				mesh_i = &(mesh_i->child());

				// adapt problem
				fmg::Adaptor::adapt(*problem_i, *mesh_i);
				problem_i = &(problem_i->child());
				space_i = problem_i->test_space().get();
			}

			// set function space
			levels[i]->space = reference_to_no_delete_pointer(*space_i);

			// init solution vector (may be omitted)
			std::size_t n = space_i->dim();
			levels[i]->x->resize(n);

			if (i > 0) {
				// init residual/working vector (may be omitted)
				levels[i]->r->resize(n);

				// init prolongation
				levels[i-1]->init_prolongation();
			}
		}

		// get coarse and fine level
		boost::shared_ptr<mg_level> c_level = levels[0];
		boost::shared_ptr<mg_level> f_level = levels[levels.size()-1];

		// collect Dirichlet boundary conditions
		std::vector<boost::shared_ptr<const BoundaryCondition> > bcs = problem_i->bcs();
		std::vector<const DirichletBC*> dbcs;
		for (std::size_t k = 0; k < bcs.size(); k++) {
			const DirichletBC* dbc = dynamic_cast<const DirichletBC*>(bcs[k].get());
			if (dbc == NULL) continue;
			dbcs.push_back(dbc);
		}

		// assemble linear system of finest level
		assemble_system(*(f_level->A), *(f_level->b), *(problem_i->bilinear_form()),
				*(problem_i->linear_form()), dbcs);

		// do Galerkin assembly on other levels
		for (int i = levels.size()-2; i >= 0; i--) {
			levels[i]->init_galerkin_A();
		}

		// init coarse level solver
		c_level->init_solver();

		// init smoothers
		for (std::size_t i = 1; i < levels.size(); i++) {
			levels[i]->init_smoother();
		}

		// init preconditioned cg solver
		boost::shared_ptr<const GenericLinearOperator> Ac(f_level->A.get(), NoDeleter());
		mg_prec p(f_level);
		PETScKrylovSolver solver("cg", p);
		solver.parameters["monitor_convergence"] = true;
		solver.parameters["nonzero_initial_guess"] = false;
		solver.parameters["convergence_norm_type"] = "preconditioned";
		solver.parameters["relative_tolerance"] = rtol;
		solver.parameters["absolute_tolerance"] = 0.0;
		solver.set_operator(Ac);

		Function f_u(f_level->space);

		ti.stop();

		// solve
		{
			Timer ts("solve multigrid");
			solver.solve(*(f_u.vector()), *(f_level->b));
		}

		std::cout << "dofs: " << f_u.vector()->size() << std::endl;

		//plot(f_u);
		//interactive();
	}

	list_timings(true);
	std::cout << std::endl;

	// solve problem with algebraic multigrid
	{
		std::cout << "~~~~~~~~~~~~~~~~~ AMG ~~~~~~~~~~~~~~~~~" << std::endl;

		Timer tt("total amg");
		Timer ti("init amg");

		boost::shared_ptr<LinearVariationalProblem> problem = create_problem(nrefine);
		boost::shared_ptr<Function> u = problem->solution();

		// collect Dirichlet boundary conditions
		DirichletBC::Map bcmap;
		std::vector<boost::shared_ptr<const BoundaryCondition> > bcs = problem->bcs();
		std::vector<const DirichletBC*> dbcs;
		for (std::size_t k = 0; k < bcs.size(); k++) {
			const DirichletBC* dbc = dynamic_cast<const DirichletBC*>(bcs[k].get());
			if (dbc == NULL) continue;
			dbcs.push_back(dbc);
		}

		boost::shared_ptr<GenericMatrix> A(new PETScMatrix());
		boost::shared_ptr<GenericVector> b(new PETScVector());
		assemble_system(*A, *b, *(problem->bilinear_form()), *(problem->linear_form()), dbcs);

		boost::shared_ptr<const GenericLinearOperator> Ac(A.get(), NoDeleter());
		PETScPreconditioner p("hypre_amg");
		PETScKrylovSolver solver("cg", p);
		solver.parameters["monitor_convergence"] = true;
		solver.parameters["nonzero_initial_guess"] = false;
		solver.parameters["convergence_norm_type"] = "preconditioned";
		solver.parameters["relative_tolerance"] = rtol;
		solver.parameters["absolute_tolerance"] = 0.0;
		solver.set_operator(Ac);

		ti.stop();

		// solve
		{
			Timer ts("solve amg");
			solver.solve(*(u->vector()), *b);
		}

		std::cout << "dofs: " << u->vector()->size() << std::endl;
	}

	list_timings(true);

	return 0;
}

