
// This demo program solves Poisson's equation
//
//	- div grad u(x, y) = f(x, y)
//
// on the unit square with source f given by
//
//	f(x, y) = 10*exp(-((x - 0.5)^2 + (y - 0.5)^2) / 0.02)
//
// and boundary conditions given by
//
//	u(x, y)		= 0		for x = 0 or x = 1
//	du/dn(x, y)	= sin(5*x)	for y = 0 or y = 1

#include <dolfin.h>
#include "Poisson.h"
#include <fmg.h>

using namespace dolfin;

// Source term (right-hand side)
class Source : public Expression
{
	void eval(Array<double>& values, const Array<double>& x) const
	{
		double dx = x[0] - 0.5;
		double dy = x[1] - 0.5;
		values[0] = 10*exp(-(dx*dx + dy*dy) / 0.02);
	}
};

// Normal derivative (Neumann boundary condition)
class dUdN : public Expression
{
	void eval(Array<double>& values, const Array<double>& x) const
	{
		values[0] = sin(5*x[0]);
	}
};

// Sub domain for Dirichlet boundary condition
class DirichletBoundary : public SubDomain
{
	bool inside(const Array<double>& x, bool on_boundary) const
	{
		return x[0] < DOLFIN_EPS or x[0] > 1.0 - DOLFIN_EPS;
	}
};

int main(int argc, char* argv[])
{
	// Create mesh and function space
	UnitSquareMesh mesh(7, 7);
	Poisson::FunctionSpace V(mesh);

	// Define boundary condition
	Constant u0(0.0);
	DirichletBoundary boundary;
	DirichletBC bc(V, u0, boundary);

	// Define variational forms
	Poisson::BilinearForm a(V, V);
	Poisson::LinearForm L(V);
	Source f;
	dUdN g;
	L.f = f;
	L.g = g;

	Function u(V);
	LinearVariationalProblem problem(a, L, u, bc);

	fmg::Tests tests(problem);
	tests.parameters["num_refinements"] = 5;
	tests.parameters.parse(argc, argv);
	tests.run();

	return 0;
}
