/*
 * Adaptor.cpp
 *
 *	Created on: 02.10.2012
 *      Author: Felix Ospald
 */

#include "Adaptor.h"
#include "../common/common.h"

#include <dolfin/adaptivity/adapt.h>
#include <dolfin/refinement/UniformMeshRefinement.h>
#include <dolfin/refinement/LocalMeshRefinement.h>
#include <dolfin/function/Function.h>
#include <dolfin/fem/DirichletBC.h>
#include <dolfin/fem/GenericDofMap.h>
#include <dolfin/fem/Form.h>

using namespace dolfin;

namespace fmg {

// Common function for setting parent/child
template <typename T>
void set_parent_child(const T& parent, boost::shared_ptr<T> child)
{
	// Use a const_cast so we can set the parent/child
	T& _parent = const_cast<T&>(parent);

	// Set parent/child
	_parent.set_child(child);
	child->set_parent(reference_to_no_delete_pointer(_parent));
}


// Initialize all required mesh entities of a given dofmap
void init_mesh_entites_for_dofmap(const dolfin::Mesh& adapted_mesh, const dolfin::GenericDofMap& dofmap)
{
	// Check that we have all mesh entities
	for (std::size_t d = 0; d <= adapted_mesh.topology().dim(); ++d)
	{
		if (dofmap.needs_mesh_entities(d) && adapted_mesh.num_entities(d) == 0)
			adapted_mesh.init(d);
	}
}


// Initialize all required mesh entities for each dofmap of the coefficient function spaces of a given form
void init_mesh_entites_for_dofmap(const dolfin::Mesh& adapted_mesh, const dolfin::Form& form)
{
	std::vector<boost::shared_ptr<const dolfin::FunctionSpace> > spaces = form.function_spaces();

	for (std::size_t i = 0; i < spaces.size(); i++) {
		init_mesh_entites_for_dofmap(adapted_mesh, *(spaces[i]->dofmap()));
	}

	std::vector<boost::shared_ptr<const GenericFunction> > coefs = form.coefficients();

	for (std::size_t i = 0; i < coefs.size(); i++)
	{
	    const Function* f = dynamic_cast<const Function*>(coefs[i].get());

	    if (f) {
	    	init_mesh_entites_for_dofmap(adapted_mesh, *(f->function_space()->dofmap()));
	    }
	}

}


// Adapt mesh uniformly.
const dolfin::Mesh& Adaptor::adapt(const dolfin::Mesh& mesh)
{
	FMGTimer t("adapt_mesh");

	boost::shared_ptr<Mesh> adapted_mesh(new Mesh());
	dolfin::UniformMeshRefinement::refine(*adapted_mesh, mesh);
	set_parent_child(mesh, adapted_mesh);
/*
	// Initialize the entities initialized in mesh in adapted_mesh.
	// This is not a good idea, since not all entities are used always!
	// The function which needs the entities should initialize them.
	for (std::size_t d = 0; d <= mesh.topology().dim(); ++d)
		if (mesh.num_entities(d) != 0)
			adapted_mesh->init(d);
*/
	return *adapted_mesh;
}


// Adapt marked cells of mesh.
const dolfin::Mesh& Adaptor::adapt(
	const dolfin::Mesh& mesh,
	const dolfin::MeshFunction<bool>& cell_markers)
{
	FMGTimer t("adapt_mesh");

	boost::shared_ptr<Mesh> adapted_mesh(new Mesh());
	dolfin::LocalMeshRefinement::refine(*adapted_mesh, mesh, cell_markers);
	set_parent_child(mesh, adapted_mesh);
/*
	// Initialize the entities initialized in mesh in adapted_mesh.
	// This is not a good idea, since not all entities are used always!
	// The function which needs the entities should initialize them.
	for (std::size_t d = 0; d <= mesh.topology().dim(); ++d)
		if (mesh.num_entities(d) != 0)
			adapted_mesh->init(d);
*/
	return *adapted_mesh;
}


// Adapt linear variational problem.
const dolfin::LinearVariationalProblem& Adaptor::adapt(
	const dolfin::LinearVariationalProblem& problem,
	const dolfin::Mesh& adapted_mesh,
	bool interpolate_coef, bool interpolate_solution)
{
	return adapt(problem, reference_to_no_delete_pointer(adapted_mesh), interpolate_coef, interpolate_solution);
}


// Adapt linear variational problem.
const dolfin::LinearVariationalProblem& Adaptor::adapt(
	const dolfin::LinearVariationalProblem& problem,
	boost::shared_ptr<const dolfin::Mesh> adapted_mesh,
	bool interpolate_coef, bool interpolate_solution)
{
	FMGTimer t("adapt_problem");

	// The following code is identical to the adapt function for LinearVariationalProblem in
	// dolfin/adaptivity/adapt.cpp
	// except that the solution variable u is not interpolated as this is not necessary
	// and may also cause extrapolation problems like
	//
	// *** Error:   Unable to evaluate function at point.
	// *** Reason:  The point is not inside the domain. Consider setting "allow_extrapolation" to allow extrapolation.
	// *** Where:   This error was encountered inside Function.cpp.

	// Skip refinement if already refined
	if (problem.has_child())
	{
		dolfin_debug("Linear variational problem has already been refined, returning child problem.");
		return problem.child();
	}

	// Get data
	boost::shared_ptr<const Form> a = problem.bilinear_form();
	boost::shared_ptr<const Form> L = problem.linear_form();
	boost::shared_ptr<const Function> u = problem.solution();
	std::vector<boost::shared_ptr<const BoundaryCondition> > bcs = problem.bcs();

	dolfin_assert(a);
	dolfin_assert(L);

	// Refine forms
	init_mesh_entites_for_dofmap(*adapted_mesh, *a); // <= changed
	dolfin::adapt(*a, adapted_mesh, interpolate_coef); // <= changed
	init_mesh_entites_for_dofmap(*adapted_mesh, *L); // <= changed
	dolfin::adapt(*L, adapted_mesh, interpolate_coef); // <= changed

	// FIXME: Note const-cast here, don't know how to get around it

	// Refine solution variable
	dolfin_assert(u);
	dolfin::adapt(*u, adapted_mesh, interpolate_solution); // <= changed
	boost::shared_ptr<Function> refined_u =
		reference_to_no_delete_pointer(const_cast<Function&>(u->child()));

	// Refine bcs
	boost::shared_ptr<const FunctionSpace> V(problem.trial_space());
	std::vector<boost::shared_ptr<const BoundaryCondition> > refined_bcs;
	for (std::size_t i = 0; i < bcs.size(); i++)
	{
		const DirichletBC* bc = dynamic_cast<const DirichletBC*>(bcs[i].get());
		if (bc != 0)
		{
			dolfin_assert(V);
			dolfin::adapt(*bc, adapted_mesh, *V);
			refined_bcs.push_back(bc->child_shared_ptr());
		}
		else
		{
			dolfin_error("adapt.cpp",
					"adapt linear variational problem",
					"Only implemented for Dirichlet boundary conditions");
		}
	}

	// Create new problem
	dolfin_assert(a);
	dolfin_assert(L);
	dolfin_assert(u);
	boost::shared_ptr<LinearVariationalProblem>
		refined_problem(new LinearVariationalProblem(a->child_shared_ptr(),
		L->child_shared_ptr(),
		refined_u,
		refined_bcs));

	// Set parent / child
	set_parent_child(problem, refined_problem);

	return *refined_problem;
}


// Adapt nonlinear variational problem.
const dolfin::NonlinearVariationalProblem& Adaptor::adapt(
	const dolfin::NonlinearVariationalProblem& problem,
	const dolfin::Mesh& adapted_mesh,
	bool interpolate_coef, bool interpolate_solution)
{
	return adapt(problem, reference_to_no_delete_pointer(adapted_mesh), interpolate_coef, interpolate_solution);
}


// Adapt nonlinear variational problem.
const dolfin::NonlinearVariationalProblem& Adaptor::adapt(
	const NonlinearVariationalProblem& problem,
	boost::shared_ptr<const Mesh> adapted_mesh,
	bool interpolate_coef, bool interpolate_solution)
{
	FMGTimer t("adapt_problem");

	// Skip refinement if already refined
	if (problem.has_child())
	{
		dolfin_debug("Nonlinear variational problem has already been refined, returning child problem.");
		return problem.child();
	}

	// Get data
	boost::shared_ptr<const Form> F = problem.residual_form();
	boost::shared_ptr<const Form> J = problem.jacobian_form();
	boost::shared_ptr<const Function> u = problem.solution();
	std::vector<boost::shared_ptr<const BoundaryCondition> > bcs = problem.bcs();

	// Refine forms
	dolfin_assert(F);
	init_mesh_entites_for_dofmap(*adapted_mesh, *F); // <= changed
	dolfin::adapt(*F, adapted_mesh, interpolate_coef); // <= changed
	if (J) {
		init_mesh_entites_for_dofmap(*adapted_mesh, *J); // <= changed
		dolfin::adapt(*J, adapted_mesh, interpolate_coef); // <= changed
	}

	// FIXME: Note const-cast here, don't know how to get around it

	// Refine solution variable
	dolfin_assert(u);
	dolfin::adapt(*u, adapted_mesh, interpolate_solution); // <= changed
	boost::shared_ptr<Function> refined_u =
			reference_to_no_delete_pointer(const_cast<Function&>(u->child()));

	// Refine bcs
	boost::shared_ptr<const FunctionSpace> V(problem.trial_space());
	std::vector<boost::shared_ptr<const BoundaryCondition> > refined_bcs;
	for (std::size_t i = 0; i < bcs.size(); i++)
	{
		const DirichletBC* bc = dynamic_cast<const DirichletBC*>(bcs[i].get());
		if (bc != 0)
		{
			dolfin_assert(V);
			dolfin::adapt(*bc, adapted_mesh, *V);
			refined_bcs.push_back(bc->child_shared_ptr());
		}
		else
		{
			dolfin_error("adapt.cpp",
					   "adapt nonlinear variational problem",
					   "Only implemented for Dirichlet boundary conditions");
		}
	}

	// Create new problem
	dolfin_assert(F);
	dolfin_assert(u);
	boost::shared_ptr<NonlinearVariationalProblem> refined_problem;
	if (J)
		refined_problem.reset(new NonlinearVariationalProblem(F->child_shared_ptr(),
															  refined_u,
															  refined_bcs,
															  J->child_shared_ptr()));
	else
		refined_problem.reset(new NonlinearVariationalProblem(F->child_shared_ptr(),
															  refined_u,
															  refined_bcs));

	// Set parent / child
	set_parent_child(problem, refined_problem);

	return *refined_problem;
}


} // namespace

