/*
 * FMGTimer.h
 *
 *  Created on: 09.11.2012
 *      Author: Felix Ospald
 */

#ifndef FMG_FMGTIMER_H_
#define FMG_FMGTIMER_H_

#include <dolfin/common/Timer.h>
#include <fmg/common/Utils.h>

namespace fmg {

/// This is an fmg specific extension to dolfin::Timer.
/// Example usage::
///
///   {
///     FMGTimer t1("task1");
///
///     // do something here
///   }
///
///   std::cout << FMGTimer::get_sum("task1") << std::endl;
///
/// .. note:: dolfin::list_timings() will show all FMGTimer tasks prefixed with "fmg: ".

class FMGTimer : public dolfin::Timer
{
public:
    /// Create timer. The *task* parameter is an identifier used to collect
	/// timing information over several instance creations.
	FMGTimer(const std::string& task) : dolfin::Timer("fmg: " + task), _task(task), _stopped(false) { DEBUG_PRINT("entering " + _task); }

	/// Stops the timer and destructs the object.
	~FMGTimer() { if (!_stopped) stop(); DEBUG_PRINT("leaving " + _task); }

    /// Stop timer
    double stop();

    /// Start timer
    void start() { dolfin::Timer::start(); _stopped = false; }

    /// Clear all recorded timings.
    static void reset_all() { _tasks.clear(); }

    /// Reset the timings of *task*.
    static void reset(const std::string& task) { _tasks.erase(task); }

    /// Returns true if timings are available for *task*.
    static bool have_task(const std::string& task) { return (_tasks.count(task) > 0); }

    /// Get the number of repetitions of *task*.
    static std::size_t get_reps(const std::string& task) { return have_task(task) ? _tasks[task].n : 0; }

    /// Get the total time of *task*.
    static double get_sum(const std::string& task) { return have_task(task) ? _tasks[task].t_sum : 0; }

    /// Get the minimum time of *task*.
    static double get_min(const std::string& task) { return have_task(task) ? _tasks[task].t_min : 0; }

    /// Get the maximum time of *task*.
    static double get_max(const std::string& task) { return have_task(task) ? _tasks[task].t_max : 0; }

    /// Get the mean time of *task*.
    static double get_mean(const std::string& task) { return have_task(task) ? _tasks[task].t_mean : 0; }

protected:

    struct task_t {
    	task_t() : n(0), t_sum(0), t_min(0), t_max(0), t_mean(0) { }
    	std::size_t n;
    	double t_sum, t_min, t_max, t_mean;
    };

    static task_t& get_task(const std::string& task);

    static std::map<std::string, task_t> _tasks;
    std::string _task;
    bool _stopped;

}; // class


} // namespace

#endif /* FMG_FMGTIMER_H_ */
