/*
 * PETScExt.cpp
 *
 *	Created on: 01.12.2012
 *			Author: Felix Ospald
 */

#include "PETScExt.h"

#include <petscversion.h>
#if PETSC_VERSION_MAJOR==3 && PETSC_VERSION_MINOR>2
	#include <petsc-private/pcimpl.h>
	#include <petsc-private/matimpl.h>
#else
	#include <private/pcimpl.h>
	#include <private/matimpl.h>
#endif

// this from src/mat/impls/aij/seq/aij.h
// hopefully PETSc does not change too much...

#define SEQAIJHEADER(datatype)  \
	PetscBool         roworiented;      /* if true, row-oriented input, default */\
	PetscInt          nonew;            /* 1 don't add new nonzeros, -1 generate error on new */\
	PetscInt          nounused;         /* -1 generate error on unused space */\
	PetscBool         singlemalloc;     /* if true a, i, and j have been obtained with one big malloc */\
	PetscInt          maxnz;            /* allocated nonzeros */\
	PetscInt          *imax;            /* maximum space allocated for each row */\
	PetscInt          *ilen;            /* actual length of each row */\
	PetscBool         free_imax_ilen;  \
	PetscInt          reallocs;         /* number of mallocs done during MatSetValues() \
										as more values are set than were prealloced */\
	PetscInt          rmax;             /* max nonzeros in any row */\
	PetscBool         keepnonzeropattern;   /* keeps matrix structure same in calls to MatZeroRows()*/\
	PetscBool         ignorezeroentries; \
	PetscInt          *xtoy,*xtoyB;     /* map nonzero pattern of X into Y's, used by MatAXPY() */\
	Mat               XtoY;             /* used by MatAXPY() */\
	PetscBool         free_ij;          /* free the column indices j and row offsets i when the matrix is destroyed */ \
	PetscBool         free_a;           /* free the numerical values when matrix is destroy */ \
	Mat_CompressedRow compressedrow;    /* use compressed row format */                      \
	PetscInt          nz;               /* nonzeros */                                       \
	PetscInt          *i;               /* pointer to beginning of each row */               \
	PetscInt          *j;               /* column values: j + i[k] - 1 is start of row k */  \
	PetscInt          *diag;            /* pointers to diagonal elements */                  \
	PetscBool         free_diag;         \
	datatype          *a;               /* nonzero elements */                               \
	PetscScalar       *solve_work;      /* work space used in MatSolve */                    \
	IS                row, col, icol;   /* index sets, used for reorderings */ \
	PetscBool         pivotinblocks;    /* pivot inside factorization of each diagonal block */ \
	Mat               parent             /* set if this matrix was formed with MatDuplicate(...,MAT_SHARE_NONZERO_PATTERN,....);
										 means that this shares some data structures with the parent including diag, ilen, imax, i, j */
typedef struct {
	SEQAIJHEADER(MatScalar);
} Mat_SeqAIJ;


namespace fmg {

PetscErrorCode PETScExt::MatZeroRowsColumnsIS_SeqAIJ(Mat mat, IS isrows, IS iscols)
{
	PetscErrorCode ierr;
	PetscInt numRows;
	PetscInt numCols;
	const PetscInt *rows;
	const PetscInt *cols;

	PetscFunctionBegin;
	PetscValidHeaderSpecific(mat,MAT_CLASSID,1);
	PetscValidHeaderSpecific(isrows,IS_CLASSID,2);
	PetscValidHeaderSpecific(iscols,IS_CLASSID,2);
	PetscValidType(mat,1);
	PetscValidType(isrows,2);
	PetscValidType(iscols,2);
	ierr = ISGetLocalSize(isrows,&numRows);CHKERRQ(ierr);
	ierr = ISGetIndices(isrows,&rows);CHKERRQ(ierr);
	ierr = ISGetLocalSize(iscols,&numCols);CHKERRQ(ierr);
	ierr = ISGetIndices(iscols,&cols);CHKERRQ(ierr);
	ierr = MatZeroRowsColumns_SeqAIJ(mat,numRows,rows,numCols,cols);CHKERRQ(ierr);
	ierr = ISRestoreIndices(isrows,&rows);CHKERRQ(ierr);
	ierr = ISRestoreIndices(iscols,&cols);CHKERRQ(ierr);
	PetscFunctionReturn(0);
}


PetscErrorCode PETScExt::MatZeroRowsColumns_SeqAIJ(Mat A, PetscInt M, const PetscInt rows[], PetscInt N, const PetscInt cols[])
{
	Mat_SeqAIJ *a = (Mat_SeqAIJ*)A->data;
	PetscInt i,j,m,n;
	PetscErrorCode ierr;
	PetscBool *zeroed;

	PetscFunctionBegin;

	MatGetSize(A, &m, &n);

	for (i=0; i<M; i++) {
		if (rows[i] < 0 || rows[i] >= m) SETERRQ1(PETSC_COMM_SELF,PETSC_ERR_ARG_OUTOFRANGE,"row %D out of range", rows[i]);
		ierr = PetscMemzero(&a->a[a->i[rows[i]]],a->ilen[rows[i]]*sizeof(PetscScalar));CHKERRQ(ierr);
	}

	if (N > 0)
	{
		ierr = PetscMalloc(n*sizeof(PetscBool),&zeroed);CHKERRQ(ierr);
		ierr = PetscMemzero(zeroed,n*sizeof(PetscBool));CHKERRQ(ierr);

		for (i=0; i<N; i++) {
			if (cols[i] < 0 || cols[i] >= n) SETERRQ1(PETSC_COMM_SELF,PETSC_ERR_ARG_OUTOFRANGE,"column %D out of range", cols[i]);
			zeroed[cols[i]] = PETSC_TRUE;
		}

		// this is expensive, but better than nothing...
		for (i=0; i<A->rmap->n; i++) {
			for (j=a->i[i]; j<a->i[i+1]; j++) {
				if (zeroed[a->j[j]]) {
					a->a[j] = 0.0;
				}
			}
		}

		ierr = PetscFree(zeroed);CHKERRQ(ierr);
	}

	A->same_nonzero = PETSC_TRUE;
	ierr = MatAssemblyEnd(A,MAT_FINAL_ASSEMBLY);CHKERRQ(ierr);
	PetscFunctionReturn(0);
}


} // namespace
