/*
 * VectorCyclePattern.h
 *
 *  Created on: 02.10.2012
 *      Author: Felix Ospald
 */

#ifndef FMG_VECTORCYCLEPATTERN_H_
#define FMG_VECTORCYCLEPATTERN_H_

#include <fmg/cycle/GenericCyclePattern.h>

#include <vector>
#include <string>

namespace fmg {

/// Create cycle pattern from vector by specifying number of cycles per level.
///
/// Examples for the initialization vector cycles_per_level:
///
///   =======  ================================
///   Pattern  Description
///   =======  ================================
///   <1>      direct solve only
///   <1,1>    2-grid V-cycle
///   <1,1,1>  3-grid V-cycle
///   <2,1,1>  3-grid VV-cycle
///   <1,2,1>  3-grid W-cycle
///   <2,2,1>  3-grid WW-cycle
///   =======  ================================
///
/// Examples for concatenated cycle patterns (vec_cycles_per_level):
///
///   =====================  ===================================
///   Vector                 Description
///   =====================  ===================================
///   <<1,1,1>,<1,2,1>>      3-grid VW-cycle
///   <<1,2,1,1>,<1,1,1,1>>  4-grid wV-cycle (w: 3-grid W-cycle)
///   =====================  ===================================
///

class VectorCyclePattern : public GenericCyclePattern
{
public:
	/// Create empty cycle pattern. Use add_pattern(...) for initialization in derived class.
	VectorCyclePattern() { reset(); }

	/// Create cycle pattern from vector by specifying number of cycles per level.
	VectorCyclePattern(const std::vector<std::size_t>& cycles_per_level);

	/// Create concatenated cycle pattern from vector of vectors each specifying number of cycles per level.
	VectorCyclePattern(const std::vector<std::vector<std::size_t> >& vec_cycles_per_level);

	/// Create cycle pattern from string of concatenated numbers of cycles per level separated by any non-numeric characters,
	/// except for the following:
	///
	///   | '-' : finishes the current pattern and concatenates the pattern to the previous and starts a new pattern
	///   | '*' : the current pattern is padded with 1-cycles such that the pattern has number_of_levels levels and do the same as in '-'.
	///   | 'V' : substitute for '*'
	///
	/// Example: pattern="2,1" or "2*" or "1*1*" or "1 1-1*" or "1e1-1x1" or "VV" or "**" and number_of_levels=3 gives 3-grid VV-cycle.
	VectorCyclePattern(const std::string& pattern, std::size_t number_of_levels);

	void reset() { _counter = 0; _cycle = 0; }
	void next() { _counter++; if (action_index() == 0) _cycle++; }
	Action action() const { return _pattern[action_index()].action; }
	std::size_t action_index() const { return (_counter % _pattern.size()); }
	std::size_t cycle_index() const { return _cycle; }
	std::size_t level_index() const { return _pattern[action_index()].level; }

protected:
	// You may use this routine to concatenate different patterns (i.e. for VW or WV pattern)
	void add_pattern(const std::vector<std::size_t>& cycles_per_level);
	void add_pattern(const std::vector<std::size_t>& cycles_per_level, std::size_t level, std::size_t level_offset);
	// You need to call this after the last add_pattern call.
	void close_pattern();

	typedef struct {
		int level;
		Action action;
	} PatternInfo;

	std::vector<PatternInfo> _pattern;
	std::size_t _counter;
	std::size_t _cycle;
};

} // namespace

#endif /* FMG_VECTORCYCLEPATTERN_H_ */
