/*
 * SORSmoothingOperator.h
 *
 *  Created on: 25.10.2012
 *      Author: Felix Ospald
 */

#ifndef FMG_SORSMOOTHINGOPERATOR_H_
#define FMG_SORSMOOTHINGOPERATOR_H_

#include <fmg/relax/GenericSmoothingOperator.h>

#include <dolfin/la/GenericMatrix.h>
#include <dolfin/la/GenericVector.h>
#include <dolfin/la/PETScObject.h>

namespace fmg {

/// Implementation of SOR relaxation sweeps.

class SORSmoothingOperator : public GenericSmoothingOperator, public dolfin::PETScObject
{
public:
	typedef struct { enum e { forward, backward, symmetric }; } SORTypes;
	typedef SORTypes::e SORType;

	/// Constructor. Current valid values for the parameter type are:
	///   *SORTypes::forward*: SOR forward sweep
	///     :math:`x_{k+1} = -(D+L)^{-1}U x_k + (D+L)^{-1} b`
	///   *SORTypes::backward*: SOR backward sweep
	///     :math:`x_{k+1} = -(D+U)^{-1}L x_k + (D+U)^{-1} b`
	///   *SORTypes::symmetric*: symmetric SOR sweep
	///     :math:`x_{k+1/2} = \textrm{SOR forward sweep}(x_k)`
	/// 	:math:`x_{k+1}   = \textrm{SOR backward sweep}(x_{k+1/2})`
	///
	/// with :math:`\omega`-relaxation:
	///   *SORTypes::forward*:
	///    :math:`x_{k+1} = (D+\omega L)^{-1}(-\omega U+(1-\omega) D) x_k + \omega (D+\omega L)^{-1} b`
	///   *SORTypes::backward*:
	///    :math:`x_{k+1} = (D+\omega U)^{-1}(-\omega L+(1-\omega) D) x_k + \omega (D+\omega U)^{-1} b`
	///
	/// where :math:`A = L + D + U`.
	///
	SORSmoothingOperator(const boost::shared_ptr<const dolfin::GenericMatrix>& A, SORType type, double omega = 1.0);

	void apply(dolfin::GenericVector& x, const dolfin::GenericVector& b, bool& x_assumed_zero, std::size_t reps) const;

protected:
	boost::shared_ptr<const dolfin::GenericMatrix> _A;
	double _omega;
	int _type;
};

} // namespace


#endif /* FMG_SORSMOOTHINGOPERATOR_H_ */
