/*
 * MultigridLevel.cpp
 *
 *  Created on: 01.10.2012
 *      Author: Felix Ospald
 */

#include "MultigridLevel.h"
#include "../common/common.h"
#include "../relax/SmoothingOperator.h"
#include "../cycle/VectorCyclePattern.h"
#include "../adapt/ProlongationAssembler.h"

#include <dolfin/la/PETScFactory.h>
#include <dolfin/function/Function.h>
#include <dolfin/fem/assemble.h>
#include <dolfin/common/timing.h>
#include <dolfin/log/log.h>

#if !(PETSC_VERSION_MAJOR == 3 && PETSC_VERSION_MINOR > 1)
#error "PETSc>3.1 required!"
#endif

namespace fmg {


void MultigridLevel::pre_smooth(std::size_t reps)
{
	if (pre_smoother.get() != NULL) {
		pre_smoother->apply(*(x), *b, x_assumed_zero, reps);
	}
}


void MultigridLevel::calc_defect()
{
	// d = A*x - b
	if (x_assumed_zero) {
		d->zero();
	}
	else {
		A->mult(*(x), *(d));
	}
	*(d) -= *(b);
}


void MultigridLevel::restrict_to_coarser_level()
{
	// coarser_level->b = P^T * d
	coarser_level->P->transpmult(*d, *(coarser_level->b));
	coarser_level->x_assumed_zero = true;
}


void MultigridLevel::coarse_solve()
{
	// x = A^-1 * b
	coarse_solver->solve(*(x), *(b));
	x_assumed_zero = false;
}


void MultigridLevel::prolongate_to_finer_level()
{
	// finer_level->e = P*x
	P->mult(*x, *(finer_level->e));
}


void MultigridLevel::correct_error()
{
	// x = x - e
	*(x) -= *(e);
}


void MultigridLevel::post_smooth(std::size_t reps)
{
	if (post_smoother.get() != NULL) {
		post_smoother->apply(*(x), *(b), x_assumed_zero, reps);
	}
}


void MultigridLevel::init_system_lhs(bool galerkin_assembly)
{
	if (galerkin_assembly && (finer_level != NULL) && finer_level->A0 && P0)
	{
		// without assemble_system (about 3x faster)
		// A0 = P0^T * finer_level->A0 * P0

		double fillratio = 1.0;

		if (!A0) {
			Mat* pA0 = new Mat();
			MatPtAP(*(dynamic_cast<dolfin::PETScMatrix*>(finer_level->A0.get()))->mat(),
					*(dynamic_cast<dolfin::PETScMatrix*>(P0.get()))->mat(),
					MAT_INITIAL_MATRIX, fillratio, pA0);
			A0.reset(new dolfin::PETScMatrix(boost::shared_ptr<Mat>(pA0, dolfin::PETScMatrixDeleter())));
		}
		else {
			Mat* pA0 = (dynamic_cast<dolfin::PETScMatrix*>(A0.get()))->mat().get();
			MatPtAP(*(dynamic_cast<dolfin::PETScMatrix*>(finer_level->A0.get()))->mat(),
					*(dynamic_cast<dolfin::PETScMatrix*>(P0.get()))->mat(),
					MAT_REUSE_MATRIX, fillratio, pA0);
		}

		if (Utils::DEBUG)
		{
			// this code prints information if Galerkin assembly method creates the same lhs as
			// assemble_system by printing the difference in linf norm

			boost::shared_ptr<dolfin::GenericMatrix> dA = FMG_FACTORY_MATRIX(linear_algebra_factory());
			dolfin::assemble(*(dA), *(a));
			*dA -= *A0;
			std::cout << "galerkin_lhs_assembly check: linf-norm of dA: " << dA->norm("linf") << std::endl;
		}
		return;
	}

	// with assemble_system

	bool reset_sparsity = !A0;
	if (!A0) A0 = FMG_FACTORY_MATRIX(linear_algebra_factory());
	dolfin::assemble(*(A0), *(a), reset_sparsity);
}


void MultigridLevel::init_system_rhs(bool galerkin_assembly)
{
	if (!b0) b0 = FMG_FACTORY_VECTOR(linear_algebra_factory());

	if (galerkin_assembly && (finer_level != NULL) && finer_level->b0 && P0)
	{
		// without assemble_system (about 3x faster)
		// b0 = P0^T * finer_level->b0

		P0->transpmult(*(finer_level->b0), *(b0));

		if (Utils::DEBUG)
		{
			// this code prints information if Galerkin assembly method creates the same rhs as
			// assemble_system by printing the difference in linf norm

			boost::shared_ptr<dolfin::GenericVector> db = FMG_FACTORY_VECTOR(linear_algebra_factory());
			dolfin::assemble(*(db), *(L));
			*db -= *b0;
			std::cout << "galerkin_rhs_assembly check: linf-norm of db: " << db->norm("linf") << std::endl;
		}
		return;
	}

	// with assemble_system

	dolfin::assemble(*(b0), *(L));
}


void MultigridLevel::init_boundary_conditions(const std::string& method)
{
	dolfin::DirichletBC::Map bcmap;
	std::vector<const dolfin::DirichletBC*> dbcs;

	// cast boundary conditions
	for (std::size_t i = 0; i < bcs.size(); i++)
	{
		const dolfin::DirichletBC* dbc = dynamic_cast<const dolfin::DirichletBC*>(bcs[i].get());

		if (dbc == NULL) {
			dolfin::dolfin_error(__FILE__,
						 "init_boudary_conditions",
						 "Boundary condition is not a dolfin::DirichletBC!");
		}

		dbcs.push_back(dbc);

		// add boundary values to map
		dbc->get_boundary_values(bcmap, method);
	}

	// create sorted version of bcmap
	std::map<std::size_t, double> sorted_bcmap(bcmap.begin(), bcmap.end());

	// save values and dof indices of map separately

	std::map<std::size_t, double>::const_iterator ibcmap;
	std::size_t ndbcs = sorted_bcmap.size();
	std::size_t i = 0;

	dbc_dofs.resize(ndbcs);
	dbc_values.resize(ndbcs);

	for (ibcmap = sorted_bcmap.begin(); ibcmap != sorted_bcmap.end(); ibcmap++) {
		// std::cout << ibcmap->first << " " <<ibcmap->second << std::endl;
		dbc_dofs[i] = ibcmap->first;
		dbc_values[i] = ibcmap->second;
		i++;
	}

	// create vector of dbc_values
	if (!dbc_values_vec) dbc_values_vec = FMG_FACTORY_VECTOR(linear_algebra_factory());
	dbc_values_vec->resize(ndbcs);
	dbc_values_vec->set(&(dbc_values[0]), ndbcs, &(Utils::ints(ndbcs)[0]));
	dbc_values_vec->apply("insert");

	// create vector x_dbc
	if (!is_linear) {
		if (!x_dbc) x_dbc = FMG_FACTORY_VECTOR(linear_algebra_factory());
		x_dbc->resize(ndbcs);
		VecGetValues(*(dynamic_cast<dolfin::PETScVector*>(u->vector().get())->vec()),
				ndbcs, reinterpret_cast<const PetscInt*>(&(dbc_dofs[0])), &(Utils::temp(ndbcs)[0]));
		x_dbc->set(&(Utils::temp(ndbcs)[0]), ndbcs, &(Utils::ints(ndbcs)[0]));
		x_dbc->apply("insert");
	}

	// create vector dbc_correct
	if (is_linear) {
		dbc_correct = dbc_values_vec;
	}
	else {
		if (!dbc_correct || (dbc_correct.get() == dbc_values_vec.get()))
			dbc_correct = FMG_FACTORY_VECTOR(linear_algebra_factory());
		dbc_correct->resize(ndbcs);
		(*dbc_correct) = (*x_dbc);
		(*dbc_correct) -= (*dbc_values_vec);
	}

	// initialize dbc mapping

	if (!global_to_reduced_dofs.size())
	{
		std::size_t n = V->dim();
		std::size_t nr = n - ndbcs;

		// init mapping vectors
		global_to_reduced_dofs.resize(n);
		reduced_to_global_dofs.resize(nr);

		// create mapping from reduced dofs to original dofs and reverse

		std::size_t r = 0, k = 0;

		for (std::size_t i = 0; i < n; i++) {
			if ((r < ndbcs) && (i == dbc_dofs[r])) {
				r++;
				global_to_reduced_dofs[i] = -1;
				continue;
			}
			global_to_reduced_dofs[i] = k;
			reduced_to_global_dofs[k] = i;
			k++;
		}

		dolfin_assert(k == nr);
	}
}


void MultigridLevel::init_reduced_system(bool update_lhs, bool update_rhs, bool update_P, bool keep_A0, bool keep_P0)
{
	bool reduce = reduce_required();
	std::size_t ndbcs = dbc_dofs.size();
	std::size_t nr = reduced_to_global_dofs.size();

	IS is_reduced_to_global_dof = NULL;
	if (reduce && ((A0 && (update_lhs || !A)) || (P0 && (update_P || !P))))
		ISCreateGeneral(PETSC_COMM_WORLD, nr, reinterpret_cast<const PetscInt*>(&(reduced_to_global_dofs[0])), PETSC_USE_POINTER, &is_reduced_to_global_dof);

	if (A0 && (update_lhs || !B)) {
		// create matrix B
		IS is = is_reduced_to_global_dof;
		IS is_dbc_dofs;

		ISCreateGeneral(PETSC_COMM_WORLD, ndbcs, reinterpret_cast<const PetscInt*>(&(dbc_dofs[0])), PETSC_USE_POINTER, &is_dbc_dofs);
		if (!reduce) ISCreateGeneral(PETSC_COMM_WORLD, A0->size(0), reinterpret_cast<const PetscInt*>(&(Utils::ints(A0->size(0))[0])), PETSC_USE_POINTER, &is);

		if (!B) {
			Mat* pB = new Mat();
			MatGetSubMatrices(*(dynamic_cast<dolfin::PETScMatrix*>(A0.get())->mat()), 1,
					&is, &is_dbc_dofs, MAT_INITIAL_MATRIX, &pB);
			B.reset(new dolfin::PETScMatrix(boost::shared_ptr<Mat>(pB, dolfin::PETScMatrixDeleter())));
		}
		else {
			Mat* pB = dynamic_cast<dolfin::PETScMatrix*>(B.get())->mat().get();
			MatGetSubMatrices(*(dynamic_cast<dolfin::PETScMatrix*>(A0.get())->mat()), 1,
					&is, &is_dbc_dofs, MAT_REUSE_MATRIX, &pB);
		}

		ISDestroy(&is_dbc_dofs);
		if (!reduce) ISDestroy(&is);
	}

	if (reduce)
	{
		if (A0 && (update_lhs || !A)) {
			// reduce A
			if (!A) {
				Mat* pA = new Mat();
				MatGetSubMatrices(*(dynamic_cast<dolfin::PETScMatrix*>(A0.get())->mat()), 1,
						 &is_reduced_to_global_dof, &is_reduced_to_global_dof, MAT_INITIAL_MATRIX, &pA);
				A.reset(new dolfin::PETScMatrix(boost::shared_ptr<Mat>(pA, dolfin::PETScMatrixDeleter())));
			}
			else {
				Mat* pA = dynamic_cast<dolfin::PETScMatrix*>(A.get())->mat().get();
				MatGetSubMatrices(*(dynamic_cast<dolfin::PETScMatrix*>(A0.get())->mat()), 1,
						 &is_reduced_to_global_dof, &is_reduced_to_global_dof, MAT_REUSE_MATRIX, &pA);
			}

			if (!keep_A0) A0.reset();
		}

		// reduce b
		if (b0 && (update_rhs || !br)) {
			if (!br) br = FMG_FACTORY_VECTOR(linear_algebra_factory());
			br->resize(nr);
			VecGetValues(*(dynamic_cast<dolfin::PETScVector*>(b0.get())->vec()),
					nr, reinterpret_cast<const PetscInt*>(&(reduced_to_global_dofs[0])), &(Utils::temp(nr)[0]));
			br->set(&(Utils::temp(nr)[0]), nr, &(Utils::ints(nr)[0]));
			br->apply("insert");
		}

		// reduce P
		if (P0 && (update_P || !P))
		{
			if (finer_level == NULL)
				dolfin::error("Having P0 without having a finer_level!");

			IS is_finer_reduced_to_global_dof;
			ISCreateGeneral(PETSC_COMM_WORLD, finer_level->reduced_to_global_dofs.size(),
					reinterpret_cast<const PetscInt*>(&(finer_level->reduced_to_global_dofs[0])), PETSC_USE_POINTER, &is_finer_reduced_to_global_dof);

			if (!P) {
				Mat* pP = new Mat();
				MatGetSubMatrices(*(dynamic_cast<dolfin::PETScMatrix*>(P0.get())->mat()), 1,
						 &is_finer_reduced_to_global_dof, &is_reduced_to_global_dof, MAT_INITIAL_MATRIX, &pP);
				P.reset(new dolfin::PETScMatrix(boost::shared_ptr<Mat>(pP, dolfin::PETScMatrixDeleter())));
			}
			else {
				Mat* pP = dynamic_cast<dolfin::PETScMatrix*>(P0.get())->mat().get();
				MatGetSubMatrices(*(dynamic_cast<dolfin::PETScMatrix*>(P0.get())->mat()), 1,
						 &is_finer_reduced_to_global_dof, &is_reduced_to_global_dof, MAT_REUSE_MATRIX, &pP);
			}

			if (!keep_P0) P0.reset();

			ISDestroy(&is_finer_reduced_to_global_dof);
		}
	}
	else // identify
	{
		if (A0 && (update_lhs || !A)) {
			if (keep_A0) {
				// hard copy matrix A0
				if (!A) A = FMG_FACTORY_MATRIX(linear_algebra_factory());
				*(A) = *(A0);
			}
			else {
				// soft copy matrix A0
				A = A0;
				A0.reset();
			}

			// set rows/cols with Dirichlet boundary to identity
			MatZeroRowsColumns(*(dynamic_cast<dolfin::PETScMatrix*>(A.get())->mat()),
				dbc_dofs.size(), reinterpret_cast<const PetscInt*>(&(dbc_dofs[0])), 1.0, NULL, NULL);
		}

		// copy rhs as shared pointer
		if (update_rhs || !br)
			br = b0;

		// copy P as shared pointer
		if (P0 && (update_P || !P))
		{
			if (finer_level == NULL)
				dolfin::error("Having P0 without having a finer_level!");

			//dolfin::error("Have no efficient way to zero rows/columns of prolongation matrix. Try disabling galerkin assembly.");

			if (keep_P0) {
				// hard copy matrix P0
				if (!P) P = FMG_FACTORY_MATRIX(linear_algebra_factory());
				*(P) = *(P0);
			}
			else {
				// soft copy matrix P0
				P = P0;
				P0.reset();
			}

			// set rows/cols with Dirichlet boundary to zero
			PETScExt::MatZeroRowsColumns_SeqAIJ(*(dynamic_cast<dolfin::PETScMatrix*>(P.get())->mat()),
					finer_level->dbc_dofs.size(), reinterpret_cast<const PetscInt*>(&(finer_level->dbc_dofs[0])),
					dbc_dofs.size(), reinterpret_cast<const PetscInt*>(&(dbc_dofs[0])));
		}
	}

	if (is_reduced_to_global_dof != NULL)
		ISDestroy(&is_reduced_to_global_dof);
}


void MultigridLevel::init_rhs()
{
	if (!br) return;

	if (!bp) bp = FMG_FACTORY_VECTOR(linear_algebra_factory());
	bp->resize(B->size(0));

	// bp = br - B*dbc_correct
	B->mult(*dbc_correct, *(bp));
	*(bp) *= -1.0;
	*(bp) += *(br);

	if (dbc_method == DirichletProjectionMethods::identify) {
		std::size_t ndbc = dbc_dofs.size();
		PetscScalar* p_dbc_correct;
		VecGetArray(*(dynamic_cast<dolfin::PETScVector*>(dbc_correct.get())->vec()), &p_dbc_correct);
		bp->set(p_dbc_correct, ndbc, &(dbc_dofs[0]));
		bp->apply("insert");
		VecRestoreArray(*(dynamic_cast<dolfin::PETScVector*>(dbc_correct.get())->vec()), &p_dbc_correct);
	}
}


void MultigridLevel::init_prolongation(const std::string& prolongation_assembly_method,
		ProlongationAssembler::BoostFlag boost_flags, double eps, bool keep_P0)
{
	if ((finer_level == NULL) || P0 || P)
		return;

	ProlongationAssembler::AssemblyMethod asm_method = ProlongationAssembler::AssemblyMethods::uniform;

	if (prolongation_assembly_method == "cellwise") {
		asm_method = ProlongationAssembler::AssemblyMethods::cellwise;
	}
	else if (prolongation_assembly_method == "dofwise") {
		asm_method = ProlongationAssembler::AssemblyMethods::dofwise;
	}

	ProlongationAssembler::DirichletProjectionMethod dbc_method = ProlongationAssembler::DirichletProjectionMethods::remove;

	if (this->dbc_method == DirichletProjectionMethods::identify) {
		dbc_method = ProlongationAssembler::DirichletProjectionMethods::zero;
	}

	if (keep_P0) {
		dbc_method = ProlongationAssembler::DirichletProjectionMethods::keep;
	}

	boost::shared_ptr<dolfin::GenericMatrix> P = ProlongationAssembler::assemble_matrix(
		*V, *(finer_level->V), linear_algebra_factory(),
		asm_method, dbc_method, boost_flags,
		&(global_to_reduced_dofs), &(reduced_to_global_dofs),
		&(finer_level->global_to_reduced_dofs), &(finer_level->reduced_to_global_dofs),
		eps);

	if (keep_P0) {
		this->P0 = P;
	}
	else {
		this->P = P;
	}
}


void MultigridLevel::init_vectors()
{
	std::size_t n = A->size(0);

	if (!b) b = FMG_FACTORY_VECTOR(linear_algebra_factory());
	b->resize(n);

	if (!x) x = FMG_FACTORY_VECTOR(linear_algebra_factory());
	x->resize(n);

	if (!d) d = FMG_FACTORY_VECTOR(linear_algebra_factory());
	d->resize(n);

	if (!e) e = FMG_FACTORY_VECTOR(linear_algebra_factory());
	e->resize(n);
}


void MultigridLevel::init_smoothers(const dolfin::Parameters& parameters)
{
	pre_smoother.reset(new SmoothingOperator(A, parameters["pre_smoother"].value_str(), parameters["pre_smoother_relax"]));

	if ((parameters["pre_smoother"].value_str() == parameters["post_smoother"].value_str()) &&
		((double)parameters["pre_smoother_relax"] == (double)parameters["post_smoother_relax"]))
	{
		post_smoother = pre_smoother;
	}
	else {
		post_smoother.reset(new SmoothingOperator(A, parameters["post_smoother"].value_str(), parameters["post_smoother_relax"]));
	}
}


void MultigridLevel::init_coarse_solver(const dolfin::Parameters& parameters)
{
	if (parameters["coarse_solver_type"].value_str() == "krylov") {
		coarse_solver = FMG_FACTORY_KRYLOV_SOLVER(linear_algebra_factory(), parameters["krylov_solver_method"], parameters["krylov_solver_preconditioner"]);
		coarse_solver->parameters.update(parameters("krylov_solver_parameters"));
	}
	else {
		coarse_solver = FMG_FACTORY_LU_SOLVER(linear_algebra_factory(), parameters["lu_solver_method"]);
		coarse_solver->parameters.update(parameters("lu_solver_parameters"));
	}

	coarse_solver->set_operator(A);
}


dolfin::GenericLinearAlgebraFactory& MultigridLevel::linear_algebra_factory() const
{
	return dolfin::PETScFactory::instance();
}


bool MultigridLevel::reduce_required() const
{
	return (dbc_method == DirichletProjectionMethods::remove) && (dbc_dofs.size() > 0);
}


boost::shared_ptr<dolfin::GenericVector> MultigridLevel::create_reduced_vector() const
{
	boost::shared_ptr<dolfin::GenericVector> v = FMG_FACTORY_VECTOR(linear_algebra_factory());
	v->resize(A->size(0));
	return v;
}


void MultigridLevel::reduce_vector(const dolfin::GenericVector& x, dolfin::GenericVector& xr) const
{
	if (!reduce_required()) {
		if(&xr != &x) xr = x;
		return;
	}

	std::size_t n = reduced_to_global_dofs.size();
	x.get_local(&(Utils::temp(n)[0]), n, &(reduced_to_global_dofs[0]));
	xr.resize(n);
	xr.set(&(Utils::temp(n)[0]), n, &(Utils::ints(n)[0]));
	xr.apply("insert");
	if (&xr == this->x.get()) x_assumed_zero = false;
}


void MultigridLevel::expand_vector(const dolfin::GenericVector& xr, dolfin::GenericVector& x) const
{
	if (!reduce_required()) {
		if(&x != &xr) x = xr;
		return;
	}

	// map to global
	std::size_t n = reduced_to_global_dofs.size();

	PetscScalar* p_xr;
	VecGetArray(*(dynamic_cast<const dolfin::PETScVector*>(&xr)->vec()), &p_xr);
	x.set(p_xr, n, &(reduced_to_global_dofs[0]));
	x.apply("insert");
	VecRestoreArray(*(dynamic_cast<const dolfin::PETScVector*>(&xr)->vec()), &p_xr);

	// set boundary values
	set_dbc_values(x);
}


void MultigridLevel::update_solution()
{
	update_solution(*u);
}


void MultigridLevel::update_solution(const dolfin::Function& u)
{
	FMG_NOT_IMPLEMENTED();
	/*
	const dolfin::GenericVector& x = *(u.vector());
	dolfin::GenericVector& tx = *(this->u->vector());

	if (&x != &tx)
		tx = x;

	set_dbc_values(tx);

	if (coarser_level != NULL) {
		coarser_level->P0->transpmult(tx, *(coarser_level->u->vector()));
		coarser_level->update_solution(*(coarser_level->u));
	}*/
}


void MultigridLevel::update_solution(const dolfin::GenericVector& xr)
{
	FMG_NOT_IMPLEMENTED();
	/*
	expand_vector(xr, *(u->vector()));

	if (coarser_level != NULL) {
		coarser_level->P->transpmult(xr, *(coarser_level->t));
		coarser_level->update_solution(*(coarser_level->t));
	}*/
}


void MultigridLevel::set_dbc_values(dolfin::GenericVector& x) const
{
	x.set(&(dbc_values[0]), dbc_dofs.size(), &(dbc_dofs[0]));
	x.apply("insert");
}


void MultigridLevel::zero_dbc_values(dolfin::GenericVector& x) const
{
	x.set(&(Utils::zero(dbc_dofs.size())[0]), dbc_dofs.size(), &(dbc_dofs[0]));
	x.apply("insert");
}

} // namespace
