/*
 * MultigridLevel.h
 *
 *  Created on: 01.10.2012
 *      Author: Felix Ospald
 */

#ifndef FMG_MULTIGRIDLEVEL_H_
#define FMG_MULTIGRIDLEVEL_H_

#include <fmg/adapt/ProlongationAssembler.h>
#include <fmg/relax/GenericSmoothingOperator.h>
#include <fmg/common/Compat.h>

#include <dolfin/la/GenericMatrix.h>
#include <dolfin/la/GenericVector.h>
#include <dolfin/la/GenericLinearSolver.h>
#include <dolfin/mesh/Mesh.h>
#include <dolfin/fem/Form.h>
#include <dolfin/fem/DirichletBC.h>

namespace fmg {

/// This class is utilized by the solver and contains all information that is needed for a multigrid level.
/// (Although not all members are needed on each level.)
///
/// .. note:: Not all members are listed in this documentation. Please view the source for more information.

class MultigridLevel : public dolfin::Variable
{
public:
	/// The index of the level (0 = finest mesh)
	std::size_t index;

	/// Pointer to finer level or NULL if this is the finest level
	boost::shared_ptr<MultigridLevel> finer_level;

	/// Pointer to coarser level or NULL if this is the coarsest level
	boost::shared_ptr<MultigridLevel> coarser_level;

	/// Flag indicating if problem is linear or nonlinear
	bool is_linear;

	/// Flag indicating if the *finer_level* is a uniform refinement of this level
	/// If this is the finest level then this flag indicates if it a uniform refinement of the *coarser_level*.
	bool is_uniform;

	/// Bilinear form of the problem (or the Jacobian form J for a nonlinear problem)
	boost::shared_ptr<const dolfin::Form> a;

	/// linear form of the problem (or the residual form F for a nonlinear problem)
	boost::shared_ptr<const dolfin::Form> L;

	/// Solution of the problem
	boost::shared_ptr<dolfin::Function> u;

	/// Function space (*L->function_space(0)*) of the problem
	boost::shared_ptr<const dolfin::FunctionSpace> V;

	/// Vector of (Dirichlet) boundary conditions of the problem
	std::vector<boost::shared_ptr<const dolfin::BoundaryCondition> > bcs;

	/// Assembled bilinear form a without Dirichlet boundary conditions applied (size not reduced)
	boost::shared_ptr<dolfin::GenericMatrix> A0;

	/// Assembled linear form L without Dirichlet boundary conditions applied (size not reduced)
	boost::shared_ptr<dolfin::GenericVector> b0;

	/// Vector of Dirichlet boundary values
	std::vector<double> dbc_values;
	boost::shared_ptr<dolfin::GenericVector> dbc_values_vec;

	/// Vector of Dirichlet boundary dof indices
	std::vector<dolfin::la_index> dbc_dofs;

	typedef struct { enum e { remove, identify }; } DirichletProjectionMethods;
	typedef DirichletProjectionMethods::e DirichletProjectionMethod;

	/// Specifies which method of reduction of the Dirichlet boundary conditions is applied
	/// Reduction means either the rows/columns are set to identity or they are physically removed from the system
	DirichletProjectionMethod dbc_method;

	/// *A0* with Dirichlet boundary dofs reduced, if *dbc_method == remove* or
	/// *A0* with Dirichlet boundary dofs set to identity, if *dbc_method == identify*.
	boost::shared_ptr<dolfin::GenericMatrix> A;

	/// *b0* with Dirichlet boundary dofs reduced, if *dbc_method == remove* or
	/// *b0* if *dbc_method == identify*.
	boost::shared_ptr<dolfin::GenericVector> br;

	/// All columns of *A0* which are Dirichlet boundary dofs (*dbc_dofs*) and either
	/// all rows if *dbc_method == identity* or
	/// all rows which are not Dirichlet boundary dofs (*[1,...,b0.size()]\dbc_dofs*) if *dbc_method == remove*
	/// This is used to compute b depending on the Dirichlet boundary values
	boost::shared_ptr<dolfin::GenericMatrix> B;

	/// *x_dbc = u->vector()* reduced by non Dirichlet dofs (only set for nonlinear problems)
	boost::shared_ptr<dolfin::GenericVector> x_dbc;

	/// *dbc_correct = dbc_values* for linear problems
	/// *dbc_correct = x_dbc - dbc_values* for nonlinear problems
	boost::shared_ptr<dolfin::GenericVector> dbc_correct;

	/// Rhs of the (reduced) problem
	/// *bp = br - B*dbc_correct* and *bp[dbc_dofs[i]] = dbc_values[i]* if *dbc_method == identity*
	boost::shared_ptr<dolfin::GenericVector> bp;

	/// Vector for mapping dofs from a reduced vector (without dbc_dofs) to the global vector (with dbc_dofs)
	/// independently of the current *dbc_method*.
	std::vector<dolfin::la_index> reduced_to_global_dofs;

	/// Vector for mapping global dofs (with dbc_dofs) to reduced dofs (without dbc_dofs)
	/// independently of the current *dbc_method*.
	/// if the dof *i* is not mapped (which means it is a Dirichlet boundary dof) then *global_to_reduced_dof[i] = -1*
	std::vector<int> global_to_reduced_dofs;

	/// The prolongation operator for interpolating vectors from this level to the finer_level (using *P*)
	/// The restriction operator for interpolating vectors from this level to the coarser_level (using adjoint of *coarser_level->P*)
	/// If Galerkin assembly is disabled P0 is not used. The prolongation operator is directly assembled as *P* instead.
	boost::shared_ptr<dolfin::GenericMatrix> P0;

	/// *P0* with Dirichlet boundary dofs reduced, if *dbc_method == remove* or
	/// *P0* with Dirichlet boundary dofs zeroed if *dbc_method == identify*.
	boost::shared_ptr<dolfin::GenericMatrix> P;

	/// The current solution (on level 0 this is the current solution of the actual problem)
	boost::shared_ptr<dolfin::GenericVector> x;

	/// Value indicating if x is assumed to be zero, this actually does not mean that *x* is zero but calculations should be performed as if x was zero
	/// If you need to ensure that *x* is zero use *x.zero()* before
	mutable bool x_assumed_zero;

	/// Rhs of current problem (i.e. = *bp* on finest level)
	/// On other levels this vector is used to store the restriction of the defect *finer_level->d* to this level
	boost::shared_ptr<dolfin::GenericVector> b;

	/// Vector containing the defect d = A*x-b
	boost::shared_ptr<dolfin::GenericVector> d;

	/// This vector is used to store the correction for *x* (i.e. the prolongation of the coarser_level->x to this level)
	/// The correction is performed as *x = x - e*
	boost::shared_ptr<dolfin::GenericVector> e;

	/// Pre- and post smoothing operators
	boost::shared_ptr<const GenericSmoothingOperator> pre_smoother;
	boost::shared_ptr<const GenericSmoothingOperator> post_smoother;

	/// Direct solver (only used on the coarse levels)
	boost::shared_ptr<dolfin::GenericLinearSolver> coarse_solver;

	// Actions used by multigrid solver
	/// Apply pre-smoother *reps* times to *x*.
	virtual void pre_smooth(std::size_t reps);
	/// Computes defect :math:`d = A x - b`.
	virtual void calc_defect();
	/// Restricts defect to coarser level: :math:`coarser\_level{\rightarrow}b = P^T d`
	virtual void restrict_to_coarser_level();
	/// Compute :math:`x = A^{-1} b`
	virtual void coarse_solve();
	/// Prolongate solution to finer level: :math:`finer\_level{\rightarrow}e = P x`
	virtual void prolongate_to_finer_level();
	/// Correct error: :math:`x = x - e`
	virtual void correct_error();
	/// Apply post-smoother *reps* times to *x*
	virtual void post_smooth(std::size_t reps);

	// Initialization helpers
	/// Assemble matrix *A0*.
	virtual void init_system_lhs(bool galerkin_assembly = true);
	/// Assemble matrix *b0*.
	virtual void init_system_rhs(bool galerkin_assembly = true);
	/// Init *dbc_values*, *dbc_values_vec*, *x_dbc*, *dbc_correct* and *dbc_dofs*.
	/// Initialize *global_to_reduced_dof* and *reduced_to_global_dof* once.
	virtual void init_boundary_conditions(const std::string& method = "topological");
	/// Create/update (if update_lhs = true) matrix *A* and *B*.
	/// Create/update (if update_rhs = true) vector *br*.
	virtual void init_reduced_system(bool update_lhs = true, bool update_rhs = true, bool update_P = true, bool keep_A0 = false, bool keep_P0 = false);
	/// Create/update vector *bp*.
	virtual void init_rhs();
	/// Initialize matrix *P0*/*P* once.
	virtual void init_prolongation(const std::string& prolongation_assembly_method = "cellwise",
			ProlongationAssembler::BoostFlag boost_flags = ProlongationAssembler::BoostFlags::space_grouping,
			double eps = 1e-13, bool keep_P0 = false);
	/// Initialize vectors *b*, *x*, *d* and *e* once.
	virtual void init_vectors();
	/// Create/update smoothers.
	virtual void init_smoothers(const dolfin::Parameters& parameters);
	/// Create/update coarse solver.
	virtual void init_coarse_solver(const dolfin::Parameters& parameters);

	/// Returns a reference to the linear algebra factory used to create matrices/vectors.
	dolfin::GenericLinearAlgebraFactory& linear_algebra_factory() const;

	/// Returns true if *solve()* requires a reduced vector.
	bool reduce_required() const;

	/// Create a vector suitable for *solve()* (vector without boundary dofs)
	boost::shared_ptr<dolfin::GenericVector> create_reduced_vector() const;

	/// Copy vector *x* to *xr* with boundary dofs removed.
	void reduce_vector(const dolfin::GenericVector& x, dolfin::GenericVector& xr) const;

	/// Copy vector *xr* to *x* with boundary values inserted.
	void expand_vector(const dolfin::GenericVector& xr, dolfin::GenericVector& x) const;

	/// Update the solution given by *u* to all coarser levels (not implemented yet).
	void update_solution();

	/// Update the solution of the problem to function *u* to all coarser levels (not implemented yet).
	void update_solution(const dolfin::Function& u);

	/// Update the solution of the problem to reduced vector *xr* to all coarser levels (not implemented yet).
	void update_solution(const dolfin::GenericVector& xr);

	/// Set *x[dbc_dofs[i]] = dbc_values[i]*.
	void set_dbc_values(dolfin::GenericVector& x) const;

	/// Set *x[dbc_dofs[i]] = 0*.
	void zero_dbc_values(dolfin::GenericVector& x) const;

}; // class

} // namespace

#endif /* FMG_MULTIGRIDLEVEL_H_ */
