/*
 * MultigridPreconditionedKrylovSolver.h
 *
 *  Created on: 01.10.2012
 *      Author: Felix Ospald
 */

#ifndef FMG_MULTIGRIDPRECONDITIONEDKRYLOVSOLVER_H_
#define FMG_MULTIGRIDPRECONDITIONEDKRYLOVSOLVER_H_

#include <fmg/solve/MultigridPreconditioner.h>

namespace fmg {

/// This class provides a PETScKrylovSolver which uses a MultigridSolver as preconditioner.
/// Example usage for 3 refinements:
///
/// .. code-block:: c
///
///   // ... create coarse bilinearform a, linearform L, solution u and boundary conditions bcs
///
///   LinearVariationalProblem problem(a, L, u, bcs);
///
///   // create multigrid solver
///   fmg::MultigridSolver mg_solver(problem, 3);
///
///   // set some parameters (optional)
///   mg_solver.parameters["prolongation_assembly_mode"] = "uniform";
///   mg_solver.parameters["direct_solver_type"] = "lu";
///
///   // create multigrid pcg solver
///   fmg::MultigridPreconditionedKrylovSolver mgpcg_solver(mg_solver, "cg");
///
///   // solve problem
///   mgpcg_solver.solve();
///
///   // plot solution
///   plot(mgpcg_solver.solution());
///

class MultigridPreconditionedKrylovSolver : public dolfin::PETScKrylovSolver, public MultigridPreconditioner
{
public:
	/// Constructor. Note the passed mg_sover is modified to work as a preconditioner i.e.:
	///
	/// .. code-block:: c
	///
	///   mg_solver.parameters["error_on_nonconvergence"] = false;
	///   mg_solver.parameters["tolerance_checking"] = false;
	///   mg_solver.parameters["maximum_iterations"] = 1;
	///
    /// are set.
	///
	/// The method sting can be any valid value as for PETScKrylovSolver.
	/// The "default" method is "cg" for symmetric problems and "gmres" for non-symmetric problems.
	///
	MultigridPreconditionedKrylovSolver(MultigridSolver& mg_solver, const std::string& method = "default");

	/// Construct from given multigird problem info.
	MultigridPreconditionedKrylovSolver(MultigridProblem& problem, const dolfin::Parameters& mg_params = dolfin::empty_parameters, const std::string& method = "default");

	/// Construct from given linear problem (adaptively refined).
	/// If *nrefine* is greater zero, the problem will be refined uniformly *nrefine* times.
	/// If the problem was adapted non-uniformly in advance you have to set *uniform* to *false*.
	/// Parameters for the multigrid solver can be passed with *mg_params*.
	/// For non-symmetric problems set *symmetric* explicitly to *false*.
	/// The "default" method is "cg" for symmetric problems and "gmres" for non-symmetric problems.
	MultigridPreconditionedKrylovSolver(dolfin::LinearVariationalProblem& problem,
		std::size_t nrefine = 0, bool symmetric = true, bool uniform = true,
		bool interpolate_coef = false, bool interpolate_solution = false,
		const dolfin::Parameters& mg_params = dolfin::empty_parameters, const std::string& method = "default");

	/// Construct from given nonlinear problem (adaptively refined).
	/// If *nrefine* is greater zero, the problem will be refined uniformly *nrefine* times.
	/// If the problem was adapted non-uniformly in advance you have to set *uniform* to *false*.
	/// Parameters for the multigrid solver can be passed with *mg_params*.
	/// For symmetric problems set *symmetric* explicitly to *true*.
	/// The "default" method is "cg" for symmetric problems and "gmres" for non-symmetric problems.
	MultigridPreconditionedKrylovSolver(dolfin::NonlinearVariationalProblem& problem,
		std::size_t nrefine = 0, bool symmetric = false, bool uniform = true,
		bool interpolate_coef = false, bool interpolate_solution = false,
		const dolfin::Parameters& mg_params = dolfin::empty_parameters, const std::string& method = "default");

	/// Initialize the solver.
	void init() { _mg_solver->init(); }

	/// Return true if solver is initialized.
	bool initialized() const { return _mg_solver->initialized(); }

	/// Solve the problem (as defined in the mg_solver)
	/// The solution may be obtained with *mg_solver().solution()*
	std::size_t solve(bool x_assumed_zero = true);

	/// Solve the problem (as defined in the mg_solver)
	/// The solution is stored to *x*
	std::size_t solve(dolfin::GenericVector& x, bool x_assumed_zero = true);

	/// Solve the problem (as defined in the mg_solver)
	/// The solution is stored to *x*
	std::size_t solve(dolfin::GenericVector& x, const dolfin::GenericVector& b, bool x_assumed_zero);
	std::size_t solve(dolfin::GenericVector& x, const dolfin::GenericVector& b) { return solve(x, b, true); }

	/// Return solution *u* of the problem, when solved without passing a vector to *solve()*.
	const boost::shared_ptr<dolfin::Function> solution() const { return _mg_solver->solution(); }

protected:
	void init(const dolfin::Parameters& mg_params);
	static std::string decide_method(const std::string& method, bool symmetric);

	boost::shared_ptr<dolfin::GenericVector> _xr;

}; // class

} // namespace

#endif /* FMG_MULTIGRIDPRECONDITIONEDKRYLOVSOLVER_H_ */
