/*
 * MultigridProblem.h
 *
 *  Created on: 01.10.2012
 *      Author: Felix Ospald
 */

#ifndef FMG_MULTIGRIDPROBLEM_H_
#define FMG_MULTIGRIDPROBLEM_H_

#include <fmg/solve/MultigridLevel.h>

#include <dolfin/common/Variable.h>
#include <dolfin/fem/LinearVariationalProblem.h>
#include <dolfin/fem/NonlinearVariationalProblem.h>

namespace fmg {

/// This class is utilized by the solver and contains all information that is needed for multigrid.

class MultigridProblem : public dolfin::Variable
{
public:
	/// Construct from given linear problem (adaptively refined).
	/// If *nrefine* is greater zero, the problem will be refined uniformly *nrefine* times.
	/// For non-symmetric problems set *symmetric* explicitly to *false*.
	/// The *uniform* flag indicates if *problem* (and its children) was uniformly refined or not.
	MultigridProblem(dolfin::LinearVariationalProblem& problem, std::size_t nrefine = 0, bool symmetric = true, bool uniform = true,
			bool interpolate_coef = false, bool interpolate_solution = false);

	/// Construct from given nonlinear problem (adaptively refined).
	/// If *nrefine* is greater zero, the problem will be refined uniformly *nrefine* times.
	/// The *uniform* flag indicates if *problem* (and its children) was uniformly refined or not.
	/// For symmetric problems set *symmetric* explicitly to *true*.
	MultigridProblem(dolfin::NonlinearVariationalProblem& problem, std::size_t nrefine = 0, bool symmetric = false, bool uniform = true,
			bool interpolate_coef = false, bool interpolate_solution = false);

	/// Refine problem uniformly.
	/// If *interpolate_coef* is set to *true* also discrete form coefficients (Functions) are interpolated to the new mesh.
	/// This is disabled by default since it does not make much sense to interpolate these values from a coarse grid.
	/// If *interpolate_solution* is set to *true* also the solution is interpolated to the new mesh.
	/// This is also disabled by default since it does not make much sense.
	void adapt(bool interpolate_coef = false, bool interpolate_solution = false);

	/// Refine problem based on cell markers.
	void adapt(const dolfin::MeshFunction<bool>& cell_markers, bool interpolate_coef = false, bool interpolate_solution = false);

	/// Refine problem based on refined mesh.
	/// The *uniform* flag indicates if the *refined_mesh* is a uniform refinement of the mesh of the *finest_level()*.
	void adapt(const dolfin::Mesh& refined_mesh, bool uniform = false, bool interpolate_coef = false, bool interpolate_solution = false);

	/// Refine problem based on refined mesh (shared_ptr version).
	void adapt(const boost::shared_ptr<const dolfin::Mesh> & refined_mesh, bool uniform = false, bool interpolate_coef = false, bool interpolate_solution = false);

	/// Return solution *u* of the problem.
	const boost::shared_ptr<dolfin::Function> solution() const { return finest_level().u; }

	/// Return i-th level of the problem.
	MultigridLevel& level(std::size_t i) const { return *(_levels[i]); }

	/// Return finest level of the problem.
	MultigridLevel& finest_level() const { return *(_levels[_levels.size()-1]); }

	/// Return number of levels.
	std::size_t num_levels() const { return _levels.size(); }

	/// Return vector of levels.
	const std::vector<boost::shared_ptr<MultigridLevel> > levels() const { return _levels; }

	/// Indicates whether the problem is linear or not.
	bool is_linear() const { return _is_linear; }

	/// Indicates whether the problem is symmetric or not.
	bool is_symmetric() const { return _is_symmetric; }

	/// Indicates whether the problem is uniformly refined or not.
	bool is_uniform() const { return _is_uniform; }

protected:
	/// User may return a new instance of a custom subclass of MultigridLevel.
	virtual MultigridLevel* create_level() const;

	/// Vector of multigrid level info
	std::vector<boost::shared_ptr<MultigridLevel> > _levels;

	/// Flag indicating if problem is nonlinear.
	bool _is_linear;

	/// Flag indicating if problem is symmetric.
	bool _is_symmetric;

	/// Flag indicating if problem is uniformly refined.
	bool _is_uniform;

private:
	template<typename T>	void add_problem(T& problem, bool uniform);
	template<typename T> void add_problems(T& problem, bool uniform);

}; // class

} // namespace

#endif /* FMG_MULTIGRIDPROBLEM_H_ */
