/*
 * MultigridSolver.cpp
 *
 *  Created on: 01.10.2012
 *      Author: Felix Ospald
 */

#include "MultigridSolver.h"
#include "../common/common.h"
#include "../cycle/VectorCyclePattern.h"

#include <dolfin/la/PETScFactory.h>

#include <float.h>
#include <limits.h>
#include <iomanip>

namespace fmg {

MultigridSolver::MultigridSolver(MultigridProblem& problem)
{
	init_with_problem(dolfin::reference_to_no_delete_pointer(problem));
}


MultigridSolver::MultigridSolver(dolfin::LinearVariationalProblem& problem, std::size_t nrefine,
		bool symmetric, bool uniform, bool interpolate_coef, bool interpolate_solution)
{
	init_with_problem(boost::shared_ptr<MultigridProblem>(new MultigridProblem(problem, nrefine,
			symmetric, uniform, interpolate_coef, interpolate_solution)));
}


MultigridSolver::MultigridSolver(dolfin::NonlinearVariationalProblem& problem, std::size_t nrefine,
		bool symmetric, bool uniform, bool interpolate_coef, bool interpolate_solution)
{
	init_with_problem(boost::shared_ptr<MultigridProblem>(new MultigridProblem(problem, nrefine,
			symmetric, uniform, interpolate_coef, interpolate_solution)));
}


void MultigridSolver::init_with_problem(const boost::shared_ptr<MultigridProblem>& problem)
{
	_problem = problem;
	_initialized = false;
	parameters = default_parameters(problem->is_linear(), problem->is_symmetric(), problem->is_uniform());
}


void MultigridSolver::init()
{
	if (_initialized) {
		dolfin::warning("Solver already initialized!");
		return;
	}

	FMGTimer t("init");

	_initialized = true;
	update();
}


void MultigridSolver::update(UpdateFlag flags)
{
	int num_levels = _problem->num_levels();

	for (int i = num_levels-1; i >= 0; i--)
		update_level(i, flags);
}


void MultigridSolver::update_level(std::size_t i, UpdateFlag flags)
{
	FMGTimer t("update_level");

	const std::vector<boost::shared_ptr<MultigridLevel> >& levels = _problem->levels();

	if (!levels[i]->A) {
		MultigridLevel::DirichletProjectionMethod dbc_method = MultigridLevel::DirichletProjectionMethods::remove;

		if (parameters["dirichlet_projection_method"].value_str() == "identify")
			dbc_method = MultigridLevel::DirichletProjectionMethods::identify;

		levels[i]->dbc_method = dbc_method;
	}

	if (flags & UpdateFlags::boundary_conditions)
		init_boundary_conditions(*(levels[i]));

	if (flags & UpdateFlags::prolongation)
		init_prolongation(*(levels[i]));

	//if (!is_linear()) {
	//	update_solution();
	//}

	if (flags & UpdateFlags::system_lhs) {
		if (levels[i]->finer_level == NULL || !(bool)parameters["galerkin_lhs_assembly"]) {
			init_system_lhs(*(levels[i]));
		}
		if (levels[i]->coarser_level != NULL && (bool)parameters["galerkin_lhs_assembly"]) {
			if (!levels[i]->coarser_level->P0) init_prolongation(*(levels[i]->coarser_level));
			if (!levels[i]->coarser_level->A) init_system_lhs(*(levels[i]->coarser_level));
		}
	}

	if (flags & UpdateFlags::system_rhs) {
		if (levels[i]->finer_level == NULL || !(bool)parameters["galerkin_rhs_assembly"]) {
			init_system_rhs(*(levels[i]));
		}
		if (levels[i]->coarser_level != NULL && (bool)parameters["galerkin_rhs_assembly"]) {
			if (!levels[i]->coarser_level->P0) init_prolongation(*(levels[i]->coarser_level));
			if (!levels[i]->coarser_level->b0) init_system_rhs(*(levels[i]->coarser_level));
		}
	}

	if (flags & (UpdateFlags::system_lhs | UpdateFlags::system_rhs | UpdateFlags::prolongation))
		init_reduced_system(*(levels[i]), flags & UpdateFlags::system_lhs, flags & UpdateFlags::system_rhs, flags & UpdateFlags::prolongation);

	if (flags & UpdateFlags::system_rhs)
		init_rhs(*(levels[i]));

	if (flags & UpdateFlags::vectors)
		init_vectors(*(levels[i]));

	if (flags & UpdateFlags::smoothers)
		init_smoothers(*(levels[i]));

	if (flags & UpdateFlags::user)
		init_user(*(levels[i]));
}


std::size_t MultigridSolver::solve(bool x_assumed_zero)
{
	if (!_initialized) init();

	MultigridLevel& flevel = finest_level();
	std::size_t r = solve(*(flevel.x), x_assumed_zero);
	flevel.expand_vector(*(flevel.x), *(flevel.u->vector()));
	return r;
}


std::size_t MultigridSolver::solve(dolfin::Function& u, bool u_assumed_zero)
{
	if (!_initialized) init();

	boost::shared_ptr<dolfin::GenericVector> x = create_reduced_vector();
	std::size_t r = solve(*x, u_assumed_zero);
	expand_vector(*x, *(u.vector()));
	return r;
}


std::size_t MultigridSolver::solve(dolfin::GenericVector& xr, bool x_assumed_zero)
{
	if (!_initialized) init();

	return solve(xr, *problem_rhs(), x_assumed_zero);
}


std::size_t MultigridSolver::solve(dolfin::GenericVector& xr, const dolfin::GenericVector& br, bool x_assumed_zero)
{
	if (!_initialized) init();

	FMGTimer t("solve");
	std::size_t niter = 0;

	MultigridLevel& flevel = finest_level();
	boost::shared_ptr<dolfin::GenericVector> x_org(flevel.x);
	boost::shared_ptr<dolfin::GenericVector> b_org(flevel.b);

	try
	{
		flevel.x.reset(&xr, dolfin::NoDeleter());
		flevel.b.reset(const_cast<dolfin::GenericVector*>(&br), dolfin::NoDeleter());
		flevel.x_assumed_zero = x_assumed_zero;

		DEBUG_PRINT(" ---------- MultigridSolver::solve ----------");
		if (x_assumed_zero) DEBUG_PRINT("solve: x%d = 0", flevel.index);

		// init cycle pattern
		boost::shared_ptr<GenericCyclePattern> cycle_pattern;
		init_cylce_pattern(cycle_pattern);

		std::size_t iflevel = _problem->num_levels() - 1 - (std::size_t) parameters["level_offset"];
		bool loop = true;

		before_solve();

		do {
			std::size_t ilevel = cycle_pattern->level_index();
			MultigridLevel& level = _problem->level(iflevel - ilevel);

			if (!level.A)
				update_level(level.index);

			before_cycle_action(level, *cycle_pattern);

			switch (cycle_pattern->action()) {
				case GenericCyclePattern::Actions::restrict:
					pre_smooth_level(level);
					calc_defect_level(level);
					if (ilevel == 0) loop = loop && !test_residual_break_condition(level, *cycle_pattern);
					if (loop) restrict_to_coarser_level(level);
					break;
				case GenericCyclePattern::Actions::prolongate:
					prolongate_to_finer_level(level);
					correct_error_level(*(level.finer_level));
					post_smooth_level(*(level.finer_level));
					break;
				case GenericCyclePattern::Actions::solve:
					coarse_solve_level(level);
					if (ilevel == 0) loop = false;
					break;
				case GenericCyclePattern::Actions::pr_transition:
					pr_transition(level);
					break;
			}

			next_cycle_action(level, *cycle_pattern);

			loop = loop && !test_loop_break_condition(*cycle_pattern);
		}
		while (loop);

		after_solve();

		niter = cycle_pattern->cycle_index();
	}
	catch(...)
	{
		flevel.x = x_org;
		flevel.b = b_org;
		DEBUG_PRINT("solve: exception");
		throw;
	}

	flevel.x = x_org;
	flevel.b = b_org;

	DEBUG_PRINT("solve: end");

	return niter;
}


void MultigridSolver::set_operator(const boost::shared_ptr<const dolfin::GenericLinearOperator> A)
{
	dolfin::dolfin_error(__FILE__, "set_operator", "Method not supported.");
}


void MultigridSolver::set_operators(const boost::shared_ptr<const dolfin::GenericLinearOperator> A,
	const boost::shared_ptr<const dolfin::GenericLinearOperator> P)
{
	dolfin::dolfin_error(__FILE__, "set_operators", "Method not supported.");
}


std::size_t MultigridSolver::solve(const dolfin::GenericLinearOperator& A, dolfin::GenericVector& x, const dolfin::GenericVector& b)
{
	dolfin::dolfin_error(__FILE__, "solve(A,x,b)", "Method not supported.");
	return 0;
}


bool MultigridSolver::test_residual_break_condition(const MultigridLevel& level, const GenericCyclePattern& cycle_pattern)
{
	FMGTimer t("test_residual_break_condition");

	if ((bool)parameters["tolerance_checking"])
	{
		int cycle_index = cycle_pattern.cycle_index();
		double defect = level.d->norm(parameters["defect_norm"]);

		if (cycle_index == 0 && cycle_pattern.action_index() == 0) {
			// save first defect norm
			_defect0 = defect;
		}

		if ((bool)parameters["monitor_convergence"]) {
			dolfin::cout << "cycle " << cycle_index << ": absolute residual=" << defect << " relative residual=" << (defect/_defect0) << dolfin::endl;
		}

		if ((defect < (double)parameters["absolute_tolerance"]) || ((defect/_defect0) < (double)parameters["relative_tolerance"])) {
			return true;
		}
	}

	return false;
}


bool MultigridSolver::test_loop_break_condition(const GenericCyclePattern& cycle_pattern)
{
	FMGTimer t("test_loop_break_condition");

	int action_index = cycle_pattern.action_index();
	int cycle_index = cycle_pattern.cycle_index();

	if (action_index == 0)
	{
		// check if we exceed maximum number of cycles
		if (cycle_index >= (int)parameters["maximum_iterations"]) {
			if ((bool)parameters["error_on_nonconvergence"]) {
				dolfin::error("solution did not converge after %d iterations!", cycle_index);
			}
			return true;
		}
	}

	return false;
}


void MultigridSolver::init_boundary_conditions(MultigridLevel& level)
{
	FMGTimer t("init_boundary_conditions");

	level.init_boundary_conditions(parameters["dirichlet_dof_identification_method"]);
}


void MultigridSolver::init_reduced_system(MultigridLevel& level, bool update_lhs, bool update_rhs, bool update_P)
{
	FMGTimer t("init_reduced_system");

	level.init_reduced_system(update_lhs, update_rhs, update_P, (bool)parameters["keep_A0"], (bool)parameters["keep_P0"]);
}


void MultigridSolver::init_system_lhs(MultigridLevel& level)
{
	FMGTimer t("init_system_lhs");

	level.init_system_lhs(parameters["galerkin_lhs_assembly"]);
}


void MultigridSolver::init_system_rhs(MultigridLevel& level)
{
	FMGTimer t("init_system_rhs");

	if ((level.finer_level != NULL) && !(bool)parameters["coarse_levels_rhs_assembly"])
		return;

	level.init_system_rhs(parameters["galerkin_rhs_assembly"]);
}


void MultigridSolver::init_rhs(MultigridLevel& level)
{
	FMGTimer t("init_rhs");

	level.init_rhs();
}


void MultigridSolver::init_prolongation(MultigridLevel& level)
{
	FMGTimer t("init_prolongation");
	ProlongationAssembler::BoostFlag boost_flags = ProlongationAssembler::BoostFlags::none;

	if (parameters["prolongation_assembly_space_grouping"]) boost_flags = (ProlongationAssembler::BoostFlag)(boost_flags | ProlongationAssembler::BoostFlags::space_grouping);
	if (parameters["prolongation_assembly_hashing"]) boost_flags = (ProlongationAssembler::BoostFlag)(boost_flags | ProlongationAssembler::BoostFlags::hashing);

	level.init_prolongation(parameters["prolongation_assembly_method"], boost_flags,
			parameters["prolongation_assembly_eps"],
			((bool)parameters["galerkin_lhs_assembly"]) || ((bool)parameters["galerkin_rhs_assembly"]) || ((bool)parameters["keep_P0"]));
}


void MultigridSolver::init_vectors(MultigridLevel& level)
{
	FMGTimer t("init_vectors");

	level.init_vectors();
}


void MultigridSolver::init_smoothers(MultigridLevel& level)
{
	FMGTimer t("init_smoothers");

	level.init_smoothers(parameters);
}


void MultigridSolver::init_coarse_solver(MultigridLevel& level)
{
	FMGTimer t("init_coarse_solver");

	level.init_coarse_solver(parameters);
}


void MultigridSolver::init_user(MultigridLevel& level)
{
	FMGTimer t("init_user");
}


bool MultigridSolver::reduce_required() const
{
	return finest_level().reduce_required();
}


boost::shared_ptr<dolfin::GenericVector> MultigridSolver::create_reduced_vector() const
{
	return 	finest_level().create_reduced_vector();
}


void MultigridSolver::reduce_vector(const dolfin::GenericVector& x, dolfin::GenericVector& xr) const
{
	finest_level().reduce_vector(x, xr);
}


void MultigridSolver::expand_vector(const dolfin::GenericVector& xr, dolfin::GenericVector& x) const
{
	finest_level().expand_vector(xr, x);
}


dolfin::GenericLinearAlgebraFactory& MultigridSolver::linear_algebra_factory() const
{
	return dolfin::PETScFactory::instance();
}


void MultigridSolver::init_cylce_pattern(boost::shared_ptr<GenericCyclePattern>& cycle)
{
	int nlevels = _problem->num_levels() - (int) parameters["level_offset"];
	cycle.reset(new VectorCyclePattern(parameters["cycle_pattern"], nlevels));
}


void MultigridSolver::before_solve()
{
}


void MultigridSolver::after_solve()
{
}


void MultigridSolver::before_cycle_action(MultigridLevel& level, GenericCyclePattern& cycle)
{
}


void MultigridSolver::next_cycle_action(MultigridLevel& level, GenericCyclePattern& cycle)
{
	cycle.next();
}


void MultigridSolver::pre_smooth_level(MultigridLevel& level)
{
	FMGTimer t("pre_smooth_level");

	level.pre_smooth(parameters["pre_smoother_reps"]);
	DEBUG_PRINT("pre_smooth_level: x%d = S%d*x%d", level.index, level.index, level.index);
}


void MultigridSolver::calc_defect_level(MultigridLevel& level)
{
	FMGTimer t("calc_defect_level");

	level.calc_defect();
	DEBUG_PRINT("calc_defect: d%d = A%d*x%d - b%d", level.index, level.index, level.index, level.index);
}


void MultigridSolver::restrict_to_coarser_level(MultigridLevel& level)
{
	FMGTimer t("restrict_to_coarser_level");

	level.restrict_to_coarser_level();
	DEBUG_PRINT("restrict_to_coarser_level: b%d = P%d^T*d%d, x%d = 0", level.coarser_level->index, level.coarser_level->index, level.index, level.coarser_level->index);
}


void MultigridSolver::coarse_solve_level(MultigridLevel& level)
{
	if (level.coarse_solver.get() == NULL) {
		init_coarse_solver(level);
	}

	FMGTimer t("coarse_solve_level");
	level.coarse_solve();
	DEBUG_PRINT("coarse_solve_level: x%d = A%d^-1*b%d", level.index, level.index, level.index);
}


void MultigridSolver::prolongate_to_finer_level(MultigridLevel& level)
{
	FMGTimer t("prolongate_to_finer_level");
	dolfin_assert(!level.x_assumed_zero);
	level.prolongate_to_finer_level();
	DEBUG_PRINT("prolongate_to_finer_level: e%d = P%d*x%d", level.finer_level->index, level.index, level.index);
}


void MultigridSolver::correct_error_level(MultigridLevel& level)
{
	FMGTimer t("correct_error_level");
	dolfin_assert(!level.x_assumed_zero);
	level.correct_error();
	DEBUG_PRINT("correct_error_level: x%d = x%d - e%d", level.index, level.index, level.index);
}


void MultigridSolver::post_smooth_level(MultigridLevel& level)
{
	FMGTimer t("post_smooth_level");
	dolfin_assert(!level.x_assumed_zero);
	level.post_smooth(parameters["post_smoother_reps"]);
	DEBUG_PRINT("post_smooth_level: x%d = S%d*x%d", level.index, level.index, level.index);
}


void MultigridSolver::pr_transition(MultigridLevel& level)
{
}


const boost::shared_ptr<const dolfin::GenericMatrix> MultigridSolver::problem_operator() const
{
	return finest_level().A;
}


const boost::shared_ptr<const dolfin::GenericVector> MultigridSolver::problem_rhs() const
{
	return finest_level().bp;
}


const boost::shared_ptr<dolfin::Function> MultigridSolver::solution() const
{
	return finest_level().u;
}


void MultigridSolver::update_solution()
{
	finest_level().update_solution();
}


void MultigridSolver::update_solution(const dolfin::Function& u)
{
	finest_level().update_solution(u);
}


void MultigridSolver::update_solution(const dolfin::GenericVector& xr)
{
	finest_level().update_solution(xr);
}


MultigridLevel* MultigridSolver::create_level()
{
	return new MultigridLevel();
}


const dolfin::Parameters& MultigridSolver::default_parameters(bool linear, bool symmetric, bool uniform)
{
	typedef std::map<int, boost::shared_ptr<dolfin::Parameters> > param_map;
	static param_map _default_param_map;
	int key = (linear ? 1 : 0) & (symmetric ? 2 : 0) & (uniform ? 4 : 0);
	if (_default_param_map.count(key) > 0) return *_default_param_map[key];

	dolfin::Parameters krylov_solver_parameters(dolfin::PETScKrylovSolver::default_parameters());
	dolfin::Parameters lu_solver_parameters(dolfin::PETScLUSolver::default_parameters());

	boost::shared_ptr<dolfin::Parameters> p(new dolfin::Parameters("fmg_solver"));

	std::set<std::string> allowed_dirichlet_projection_methods;

	allowed_dirichlet_projection_methods.insert("remove");
	allowed_dirichlet_projection_methods.insert("identify");
	//allowed_dirichlet_projection_methods.insert("project");

	std::set<std::string> allowed_dirichlet_dof_identification_methods;

	allowed_dirichlet_dof_identification_methods.insert("topological");
	allowed_dirichlet_dof_identification_methods.insert("pointwise");
	allowed_dirichlet_dof_identification_methods.insert("geometric");

	std::set<std::string> allowed_smoothers;

	allowed_smoothers.insert("jacobi");
	allowed_smoothers.insert("fsor");
	allowed_smoothers.insert("bsor");
	allowed_smoothers.insert("ssor");

	std::set<std::string> allowed_prolongation_assembly_methods;

	allowed_prolongation_assembly_methods.insert("uniform");
	allowed_prolongation_assembly_methods.insert("cellwise");
	allowed_prolongation_assembly_methods.insert("dofwise");

	p->add("prolongation_assembly_method", uniform ? "uniform" : "cellwise", allowed_prolongation_assembly_methods);
	p->add("prolongation_assembly_eps", 1e-13);
	p->add("prolongation_assembly_space_grouping", true);
	p->add("prolongation_assembly_hashing", false);
	p->add("dirichlet_projection_method", linear ? "identify" : "identify", allowed_dirichlet_projection_methods);
	p->add("dirichlet_dof_identification_method", "topological", allowed_dirichlet_dof_identification_methods);
	p->add("tolerance_checking", true);
	p->add("absolute_tolerance", krylov_solver_parameters["absolute_tolerance"], 0.0, DBL_MAX);
	p->add("relative_tolerance", krylov_solver_parameters["relative_tolerance"], 0.0, DBL_MAX);
	p->add("defect_norm", "l2");
	p->add("maximum_iterations", krylov_solver_parameters["maximum_iterations"], 0, INT_MAX);
	p->add("monitor_convergence", false);
	p->add("error_on_nonconvergence", true);
	p->add("galerkin_lhs_assembly", true);
	p->add("galerkin_rhs_assembly", true);
	p->add("coarse_levels_rhs_assembly", false);
	p->add("level_offset", 0, 0, INT_MAX);
	p->add("keep_A0", false);
	p->add("keep_P0", false);

	p->add("pre_smoother", "fsor", allowed_smoothers);
	p->add("pre_smoother_relax", 1.0);
	p->add("pre_smoother_reps", 1, 0, INT_MAX);

	p->add("post_smoother", "bsor", allowed_smoothers);
	p->add("post_smoother_relax", 1.0);
	p->add("post_smoother_reps", 1, 0, INT_MAX);

	p->add("cycle_pattern", "V");

	std::set<std::string> coarse_solver_types;

	coarse_solver_types.insert("lu");
	coarse_solver_types.insert("krylov");

	p->add("coarse_solver_type", "lu", coarse_solver_types);

	std::set<std::string> krylov_solver_methods;
	std::vector<std::pair<std::string, std::string> > ks_methods = dolfin::PETScKrylovSolver::methods();

	for (std::vector<std::pair<std::string, std::string> >::iterator i = ks_methods.begin(); i != ks_methods.end(); i++) {
		krylov_solver_methods.insert(i->first);
	}

	p->add("krylov_solver_method", symmetric ? "cg" : "gmres", krylov_solver_methods);

	std::set<std::string> krylov_solver_preconditioners;
	std::vector<std::pair<std::string, std::string> > ks_preconds = dolfin::PETScKrylovSolver::preconditioners();

	for (std::vector<std::pair<std::string, std::string> >::iterator i = ks_preconds.begin(); i != ks_preconds.end(); i++) {
		krylov_solver_preconditioners.insert(i->first);
	}

	p->add("krylov_solver_preconditioner", "none", krylov_solver_preconditioners);

	krylov_solver_parameters.rename("krylov_solver_parameters");
	krylov_solver_parameters("preconditioner")["reuse"] = true;
	krylov_solver_parameters("preconditioner")["same_nonzero_pattern"] = true;
	krylov_solver_parameters["error_on_nonconvergence"] = false;
	krylov_solver_parameters["nonzero_initial_guess"] = false;
	if (krylov_solver_parameters.has_parameter("convergence_norm_type"))
		krylov_solver_parameters["convergence_norm_type"] = "preconditioned";

	p->add(krylov_solver_parameters);

	std::set<std::string> lu_solver_methods;
	std::vector<std::pair<std::string, std::string> > lus_methods = dolfin::PETScLUSolver::methods();

	for (std::vector<std::pair<std::string, std::string> >::iterator i = lus_methods.begin(); i != lus_methods.end(); i++) {
		lu_solver_methods.insert(i->first);
	}

	p->add("lu_solver_method", "petsc", lu_solver_methods);

	lu_solver_parameters.rename("lu_solver_parameters");
	lu_solver_parameters["reuse_factorization"] = true;
	lu_solver_parameters["same_nonzero_pattern"] = true;
	if (lu_solver_parameters.has_parameter("symmetric_operator"))
		lu_solver_parameters["symmetric_operator"] = symmetric;
	if (lu_solver_parameters.has_parameter("num_threads"))
		lu_solver_parameters["num_threads"] = 1;

	p->add(lu_solver_parameters);

	_default_param_map.insert(param_map::value_type(key, p));
	return *p;
}


MultigridLevel& MultigridSolver::finest_level() const
{
	/*if (!_initialized) {
		dolfin::dolfin_error(__FILE__, "finest_level",
			"Initialize solver first!");
	}*/

	int offset = (int) parameters["level_offset"];
	int nlevels = _problem->num_levels();

	if (offset >= (int) _problem->num_levels()) {
		dolfin::dolfin_error(__FILE__, "finest_level",
			"Parameter level_offset out of range!");
	}

	return _problem->level(nlevels - 1 - offset);
}


} // namespace
