/*
 * MultigridSolver.h
 *
 *  Created on: 01.10.2012
 *      Author: Felix Ospald
 */

#ifndef FMG_MULTIGRIDSOLVER_H_
#define FMG_MULTIGRIDSOLVER_H_

#include <fmg/solve/MultigridProblem.h>
#include <fmg/cycle/GenericCyclePattern.h>

#include <dolfin/parameter/Parameters.h>
#include <dolfin/fem/LinearVariationalProblem.h>
#include <dolfin/fem/NonlinearVariationalProblem.h>
#include <dolfin/function/Function.h>
#include <dolfin/la/GenericLinearSolver.h>
#include <dolfin/la/PETScObject.h>

#include <list>

namespace fmg {

/// This class provides an easy to use multigrid solver for Dolfin.
/// Example usage for 3 refinements:
///
/// .. code-block:: c
///
///   // ... create coarse bilinearform a, linearform L, solution u and boundary conditions bcs
///
///   LinearVariationalProblem problem(a, L, u, bcs);
///
///   // create multigrid solver
///   fmg::MultigridSolver mg_solver(problem, 3);
///
///   // set some parameters (optional)
///   mg_solver.parameters["prolongation_assembly_mode"] = "uniform";
///   mg_solver.parameters["coarse_solver_type"] = "lu";
///   mg_solver.parameters("lu_solver_parameters")["same_nonzero_pattern"] = true;
///
///   // solve problem
///   mg_solver.solve();
///
///   // plot solution
///   plot(mg_solver.solution());
///
/// .. note:: For performance reasons it is mostly recommended to use :cpp:class:`MultigridPreconditionedKrylovSolver`
///           instead of using this solver directly.
///
/// The following *parameters* may be configured for the solver:
///
///   ====================================== ================================ =============================================================================================================================================================================
///   Parameter                              Default                          Description
///   ====================================== ================================ =============================================================================================================================================================================
///   pre_smoother                           "fsor"                           Specifies the pre-smoother for each level. See :cpp:class:`SmoothingOperator` for a list of available smoothers.
///   pre_smoother_relax                     1.0                              Specifies the pre-smoother relaxation factor for each level.
///   pre_smoother_reps                      1                                Specifies the number of pre-smoother repetitions for each level.
///   post_smoother                          "bsor"                           Specifies the post-smoother for each level. See :cpp:class:`SmoothingOperator` for a list of available smoothers.
///   post_smoother_relax                    1.0                              Specifies the post-smoother relaxation factor for each level.
///   post_smoother_reps                     1                                Specifies the number of post-smoother repetitions for each level.
///   cycle_pattern                          "V"                              Specifies the cycle pattern. See :cpp:class:`VectorCyclePattern` for valid pattern values.
///   coarse_solver_type                     "lu"                             Specifies the type of the coarse solver. Valid values are "lu" and "krylov".
///   lu_solver_method                       "petsc"                          Specifies the name of the coarse solver when *coarse_solver_type* is set to "lu".
///   lu_solver_parameters                   Parameters                       Specifies the parameters of the coarse solver when *coarse_solver_type* is set to "lu". See Dolfin manual for valid parameters.
///   krylov_solver_method                   "cg" for symmetric problems,     Specifies the name of the Krylov solver when *coarse_solver_type* is set to "krylov".
///                                          "gmres" else
///   krylov_solver_preconditioner           "none"                           Specifies the name of the Krylov solver preconditioner when *coarse_solver_type* is set to "krylov".
///   lu_krylov_parameters                   Parameters                       Specifies the parameters of the Krylov solver when *coarse_solver_type* is set to "krylov". See Dolfin manual for valid parameters.
///   prolongation_assembly_method           "uniform" for uniform problems,  See :cpp:class:`ProlongationAssembler`.
///                                          "cellwise" else
///   prolongation_assembly_eps              1e-13                            See :cpp:class:`ProlongationAssembler`
///   prolongation_assembly_space_grouping   *true*                           See :cpp:class:`ProlongationAssembler`
///   prolongation_assembly_hashing          *false*                          See :cpp:class:`ProlongationAssembler`
///   dirichlet_projection_method            "identify"                       If set to "remove", Dirichlet boundary conditions will be removed from the linear system.
///                                                                           If set to "identify", the linear system is modified such that Dirichlet boundary conditions are identifies with the right hand side.
///   dirichlet_dof_identification_method    "topological"                    Method of Dirichlet dof identification. Valid values are: "topological", "pointwise", "geometric".
///                                                                           Refer to the Dolfin documentation for more information.
///   tolerance_checking                     *true*                           If set to *true* the *absolute_tolerance* and *relative_tolerance* values are used as stopping criterion.
///   absolute_tolerance                     1e-15                            The absolute tolerance used as stopping criterion.
///                                                                           If *tolerance_checking* is enabled the solver will stop if the current defect in the *defect_norm* norm is less than *absolute_tolerance*.
///   relative_tolerance                     1e-6                             The relative tolerance used as stopping criterion.
///                                                                           If *tolerance_checking* is enabled the solver will stop if the current defect in the *defect_norm* norm is less than *relative_tolerance* times the initial defect.
///   maximum_iterations                     10000                            The maximum number of iterations used as failure criterion.
///                                                                           If *error_on_nonconvergence* is enabled the solver will throw an exception if the number of iterations exceeds this value.
///                                                                           If *error_on_nonconvergence* is not enabled the solver will return without an error.
///   monitor_convergence                    *false*                          Print residuals for each iteration.
///   error_on_nonconvergence                *true*                           See *maximum_iterations* parameter.
///   galerkin_lhs_assembly                  *true*                           If set to *true* the lhs of the problems on the coarser levels are not assembled but computed from the lhs of the finest level using :math:`A_{l+1} = P_{l+1}^T A_l P_{l+1}`.
///   galerkin_rhs_assembly                  *true*                           If set to *true* the rhs of the problems on the coarser levels are not assembled but computed from the rhs of the finest level using :math:`b_{l+1} = P_{l+1}^T b_l`.
///   coarse_levels_rhs_assembly             *false*                          If set to *true* the rhs of the problems on the coarser levels are assembled. Usually the rhs on the coarser levels are not needed.
///   level_offset                           0                                Specifies which level is considered as the finest level for which the problem is solved.
///   keep_A0                                *false*                          If set to *true* the A0 matrix will be kept for each level.
///   keep_P0                                *false*                          If set to *true* the P0 matrix will be kept for each level.
///   ====================================== ================================ =============================================================================================================================================================================
///

class MultigridSolver : public dolfin::GenericLinearSolver, public dolfin::PETScObject
{
public:
	typedef struct { enum e {
		prolongation = 1,
		boundary_conditions = 2,
		system_lhs = 4,
		system_rhs = 8,
		vectors = 16,
		smoothers = 32,
		user = 64,
		all = 0xffff,
	}; } UpdateFlags;
	typedef UpdateFlags::e UpdateFlag;

	/// Construct from given multigird problem.
	MultigridSolver(MultigridProblem& problem);

	/// Construct from given linear problem (adaptively refined).
	/// If *nrefine* is greater zero, the problem will be refined uniformly *nrefine* times.
	/// For non-symmetric problems set *symmetric* explicitly to *false*.
	/// If the problem was adapted non-uniformly in advance you have to set *uniform* to *false*.
	MultigridSolver(dolfin::LinearVariationalProblem& problem, std::size_t nrefine = 0, bool symmetric = true, bool uniform = true,
			bool interpolate_coef = false, bool interpolate_solution = false);

	/// Construct from given nonlinear problem (adaptively refined).
	/// If *nrefine* is greater zero, the problem will be refined uniformly *nrefine* times.
	/// For symmetric problems set *symmetric* explicitly to *true*.
	/// If the problem was adapted non-uniformly in advance you have to set *uniform* to *false*.
	MultigridSolver(dolfin::NonlinearVariationalProblem& problem, std::size_t nrefine = 0, bool symmetric = false, bool uniform = true,
			bool interpolate_coef = false, bool interpolate_solution = false);

	/// Initialize the solver.
	void init();

	/// Return true if solver is initialized.
	bool initialized() const { return _initialized; }

	/// Update some aspects of the solver or the problem.
	/// This needs to be called to make changes of the parameters object take effect.
	/// The flags parameter may be a combination of bitwise ORed UpdateFlags.
	/// This is also used when solving a nonlinear problem to update the system each Newton step.
	void update(UpdateFlag flags = UpdateFlags::all);

	/// Update single level.
	void update_level(std::size_t i, UpdateFlag flags = UpdateFlags::all);

	/// Update the solution given by *get_solution()* to all coarser levels (not implemented yet).
	void update_solution();

	/// Update the solution of the problem to function *u* to all coarser levels (not implemented yet).
	void update_solution(const dolfin::Function& u);

	/// Update the solution of the problem to reduced vector *xr* to all coarser levels (not implemented yet).
	void update_solution(const dolfin::GenericVector& xr);

	// Methods for invoking the solver.
	/// Solve problem. The solution may be obtained by *get_solution()*.
	std::size_t solve(bool x_assumed_zero = true);
	/// Solve problem. The solution will be stored in *u*.
	std::size_t solve(dolfin::Function& u, bool u_assumed_zero = true);
	/// Solve problem. The solution will be stored reduced vector *xr*.
	std::size_t solve(dolfin::GenericVector& xr, bool x_assumed_zero = true);
	/// Solve problem using reduced right hand side *br*. The solution will be stored reduced vector *xr*.
	std::size_t solve(dolfin::GenericVector& xr, const dolfin::GenericVector& br, bool x_assumed_zero);

    /// Interface function for GenericLinearSolver. Calls *solve(xr, br, true)*;
    std::size_t solve(dolfin::GenericVector& xr, const dolfin::GenericVector& br) {
    	return solve(xr, br, true);
    }

	// The following methods of GenericLinearSolver are not supported:
    /// Not supported.
	void set_operator(const boost::shared_ptr<const dolfin::GenericLinearOperator> A);
    /// Not supported.
	void set_operators(const boost::shared_ptr<const dolfin::GenericLinearOperator> A,
		const boost::shared_ptr<const dolfin::GenericLinearOperator> P);
    /// Not supported.
	std::size_t solve(const dolfin::GenericLinearOperator& A, dolfin::GenericVector& x, const dolfin::GenericVector& b);

	/// Return operator *A* of the problem.
	virtual const boost::shared_ptr<const dolfin::GenericMatrix> problem_operator() const;

	/// Return vector *b* of the problem.
	virtual const boost::shared_ptr<const dolfin::GenericVector> problem_rhs() const;

	/// Return solution *u* of the problem.
	virtual const boost::shared_ptr<dolfin::Function> solution() const;

	/// User may return a reference to the linear algebra factory being used.
	/// This is currently PETSc and should not be changed.
	virtual dolfin::GenericLinearAlgebraFactory& linear_algebra_factory() const;

	/// Returns true if *solve()* requires a reduced vector.
	bool reduce_required() const;

	/// Create a vector suitable for *solve()* (vector without boundary dofs)
	boost::shared_ptr<dolfin::GenericVector> create_reduced_vector() const;

	/// Copy vector *x* to *xr* with boundary dofs removed.
	void reduce_vector(const dolfin::GenericVector& x, dolfin::GenericVector& xr) const;

	/// Copy vector *xr* to *x* with boundary values inserted.
	void expand_vector(const dolfin::GenericVector& xr, dolfin::GenericVector& x) const;

	/// Return multigrid problem object.
	const MultigridProblem& problem() const { return *_problem; }

	/// Return finest level of the problem.
	MultigridLevel& finest_level() const;

	/// Return default parameters.
	static const dolfin::Parameters& default_parameters(
			bool linear = true, bool symmetric = true, bool uniform = true);

protected:

	/// This is function is called at the beginning of solve.
	virtual void before_solve();

	/// This is function is called before return of solve.
	virtual void after_solve();

	/// User needs to return true if converged, false to continue computation or throw an exception if not converged.
	/// This is function is called only on the finest level after computing the defect.
	virtual bool test_residual_break_condition(const MultigridLevel& level, const GenericCyclePattern& cycle_pattern);

	/// User needs to return true if converged, false to continue computation or throw an exception if not converged.
	/// This is function is called at the end of every loop.
	virtual bool test_loop_break_condition(const GenericCyclePattern& cycle_pattern);

	/// User may set *level.pre_smoother* and *level.post_smoother* to modify smoothers.
	virtual void init_smoothers(MultigridLevel& level);

	/// User may set level.coarse_solver to modify coarse solver.
	/// This requires also to call *level.coarse_solver->set_operator(level.A)*
	virtual void init_coarse_solver(MultigridLevel& level);

	/// User may return a new instance of a custom subclass of *MultigridLevel*.
	virtual MultigridLevel* create_level();

	/// User may set the shared pointer to a custom cycle pattern.
	virtual void init_cylce_pattern(boost::shared_ptr<GenericCyclePattern>& cycle);

	// The following initialization methods are more complicated to re-implement.
	// Currently the calls are redirected to the corresponding level.method()
	virtual void init_system_lhs(MultigridLevel& level);
	virtual void init_system_rhs(MultigridLevel& level);
	virtual void init_boundary_conditions(MultigridLevel& level);
	virtual void init_reduced_system(MultigridLevel& level, bool update_lhs, bool update_rhs, bool update_P);
	virtual void init_rhs(MultigridLevel& level);
	virtual void init_prolongation(MultigridLevel& level);
	virtual void init_vectors(MultigridLevel& level);

	/// User defined initialization routine will be called last (does currently nothing).
	virtual void init_user(MultigridLevel& level);

	/// Called each time before a cycle action (does currently nothing).
	virtual void before_cycle_action(MultigridLevel& level, GenericCyclePattern& cycle);

	/// Called to increment the cycle action (calls cycle.next()).
	virtual void next_cycle_action(MultigridLevel& level, GenericCyclePattern& cycle);

	// The following method implement the different aspects of a basic multigrid algorithm.
	// Currently the calls are redirected to the corresponding level.method()
	virtual void pre_smooth_level(MultigridLevel& level);
	virtual void calc_defect_level(MultigridLevel& level);
	virtual void restrict_to_coarser_level(MultigridLevel& level);
	virtual void coarse_solve_level(MultigridLevel& level);
	virtual void prolongate_to_finer_level(MultigridLevel& level);
	virtual void correct_error_level(MultigridLevel& level);
	virtual void post_smooth_level(MultigridLevel& level);

	/// Called on transition between prolongation and restriction (does currently nothing).
	virtual void pr_transition(MultigridLevel& level);

	// Internally used methods.
	void init_with_problem(const boost::shared_ptr<MultigridProblem>& problem);
	void add_level(const boost::shared_ptr<const dolfin::FunctionSpace>& space);

	boost::shared_ptr<MultigridProblem> _problem;
	double _defect0;
	bool _initialized;

}; // class

} // namespace

#endif /* FMG_MULTIGRIDSOLVER_H_ */
