/*
 * Tests.cpp
 *
 *  Created on: 01.10.2012
 *      Author: Felix Ospald
 */

#include "Tests.h"
#include "../common/common.h"
#include "../cycle/VectorCyclePattern.h"
#include "../solve/MultigridPreconditionedKrylovSolver.h"

#include <dolfin.h>
#include <boost/algorithm/string.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/assign/list_of.hpp>
#include <boost/format.hpp>
#include <float.h>
#include <limits.h>

namespace fmg {

struct FindFirst {
	FindFirst(std::string i) : _i(i) { }
	std::string _i;
	bool operator()(const std::pair<std::string, std::string>& p) {
		return p.first == _i;
	}
};


class Tests::TestPattern
{
public:

	TestPattern(Table& tab, MultigridSolver& mg_solver, const dolfin::Parameters& parameters, bool test_fmg)
	{
		_tab = &tab;
		_mg_solver = dolfin::reference_to_no_delete_pointer(mg_solver);
		_parameters = dolfin::reference_to_no_delete_pointer(parameters);
		_test_fmg = test_fmg;
		_x = mg_solver.create_reduced_vector();
	}

	void add_columns()
	{
		Table& tab = *_tab;
		if (_test_fmg) {
			tab.add_int_column("iter1", "fmg iter");
			tab.add_float_column("time1", "fmg time");
		}
		tab.add_int_column("iter2", (*_parameters)["primary_solver"].value_str() + "+fmg iter");
		tab.add_float_column("time2", (*_parameters)["primary_solver"].value_str() + "+fmg time");
	}

	void run()
	{
		Table& tab = *_tab;
		double t_int = (double)(*_parameters)["integration_time"];

		if (_test_fmg)
		{
			_mg_solver->parameters["error_on_nonconvergence"] = true;
			_mg_solver->parameters["tolerance_checking"] = true;
			_mg_solver->parameters["maximum_iterations"] = 1000;

			try {
				double t_solve;
				std::size_t n = integrated_solve(*_mg_solver, "fmg", "none",
						*_x, *(_mg_solver->problem_rhs()), t_int, t_solve);
				tab["iter1"] << n;
				tab["time1"] << t_solve;
			}
			catch (...) {
				tab["iter1"] << "-";
				tab["time1"] << "-";
			}
		}

		_mg_solver->parameters["error_on_nonconvergence"] = false;
		_mg_solver->parameters["tolerance_checking"] = false;
		_mg_solver->parameters["maximum_iterations"] = 1;

		try {
			double t_solve;
			std::size_t n = integrated_solve(*_mg_solver, (*_parameters)["primary_solver"], "fmg",
					*_x, *(_mg_solver->problem_rhs()), t_int, t_solve);
			tab["iter2"] << n;
			tab["time2"] << t_solve;
		}
		catch (...) {
			tab["iter2"] << "-";
			tab["time2"] << "-";
		}
	}

protected:
	boost::shared_ptr<MultigridSolver> _mg_solver;
	boost::shared_ptr<const dolfin::Parameters> _parameters;
	boost::shared_ptr<dolfin::GenericVector> _x;
	bool _test_fmg;
	Table* _tab;
};


class TestSolver : public MultigridSolver
{
public:
	TestSolver(MultigridProblem& problem) : MultigridSolver(problem) {
		_log_residuals = false;
	}

	void log_residuals(bool enable = true)
	{
		_log_residuals = enable;
		_l2_residuals.clear();
		_linf_residuals.clear();
		_prec_l2_residuals.clear();
		_prec_linf_residuals.clear();
		_nat_residuals.clear();
	}

	void after_solve()
	{
		if (_log_residuals) {
			// cg passes r=Ax-b as rhs b, so this is ||r||_2,
			// which is referred as KSP_NORM_UNPRECONDITIONED within PETSc
			_l2_residuals.push_back(finest_level().b->norm("l2"));
			// ||r||_inf
			_linf_residuals.push_back(finest_level().b->norm("linf"));
			// after solve x contains the mg preconditioner applied to r, x=Pr, so this is ||Pr||_2,
			// which is referred as KSP_NORM_PRECONDITIONED within PETSc
			_prec_l2_residuals.push_back(finest_level().x->norm("l2"));
			// ||Pr||_inf
			_prec_linf_residuals.push_back(finest_level().x->norm("linf"));
			// sqrt(r^T P r),
			// which is referred as KSP_NORM_NATURAL within PETSc
			_nat_residuals.push_back(sqrt(finest_level().b->inner(*(finest_level().x))));
		}

		MultigridSolver::after_solve();
	}

	bool _log_residuals;
	std::vector<double> _l2_residuals;
	std::vector<double> _linf_residuals;
	std::vector<double> _prec_l2_residuals;
	std::vector<double> _prec_linf_residuals;
	std::vector<double> _nat_residuals;
};


Tests::Tests(dolfin::LinearVariationalProblem& problem, bool symmetric)
{
	init_with_problem(boost::shared_ptr<MultigridProblem>(new MultigridProblem(problem, 0, symmetric)));
}


Tests::Tests(dolfin::NonlinearVariationalProblem& problem, bool symmetric)
{
	init_with_problem(boost::shared_ptr<MultigridProblem>(new MultigridProblem(problem, 0, symmetric)));
}


void Tests::init_with_problem(const boost::shared_ptr<MultigridProblem>& problem)
{
	_problem = problem;
	parameters = default_parameters(problem->is_linear(), problem->is_symmetric(), problem->is_uniform());
}


void Tests::run()
{
	bool log_active = dolfin::LogManager::logger.is_active();
	dolfin::LogManager::logger.set_log_active(false);

	// get table format
	Table::TableFormat table_format = Table::parse_format(parameters["table_format"]);

	std::size_t nrefine = (std::size_t)parameters["num_refinements"];
	double t_int = (double)parameters["integration_time"];
	dolfin::Parameters fmg_params = parameters("fmg_solver");
	TestSolver mg_solver(*_problem);
	bool test_fmg = false;

	// init solver level 0
	FMGTimer::reset_all();
	mg_solver.parameters = fmg_params;
	mg_solver.init();

	// list initialization timings
	std::vector<double> t_loss(nrefine+1);
	std::vector<double> t_mg(nrefine+1);
	std::vector<double> t_direct(nrefine+1);
	{
		double t_init_mg, t_adapt_mesh, t_loss_mesh, t_loss_init,
			t_loss_asm_bc, t_loss_asm_lhs, t_loss_asm_rhs,
			t_loss_pro, t_loss_sr, t_loss_other;
		double init_fine_mesh_time;

		for (std::size_t i = 0; i <= nrefine; i++)
		{
			if (i > 0) {
				_problem->adapt();
				mg_solver.update_level(i);
				mg_solver.update_level(i-1, MultigridSolver::UpdateFlags::prolongation);
			}

			// we need to measure the time which is needed to create the fine mesh of the problem without adapt
			{
				const dolfin::Mesh* fine_mesh = _problem->level(i).V->mesh().get();
				boost::shared_ptr<dolfin::Mesh> equivalent_mesh;
				std::size_t num_cells = fine_mesh->num_cells();
				bool init_fine_mesh_skipped = false;

				double t0 = dolfin::time();
				std::size_t nint = 0;

				do {
					if (fine_mesh->type().dim() == 1) {
						// num_cells = nx
						std::size_t nx = num_cells;
						equivalent_mesh.reset(new dolfin::UnitIntervalMesh(nx));
					}
					else if (fine_mesh->type().dim() == 2) {
						// num_cells = 2*nx*ny
						std::size_t nx = std::max((std::size_t)1, (std::size_t) sqrt(num_cells/2));
						std::size_t ny = std::max((std::size_t)1, num_cells/(2*nx));
						equivalent_mesh.reset(new dolfin::UnitSquareMesh(nx,ny));
					}
					else if (fine_mesh->type().dim() == 3) {
						// num_cells = 6*nx*ny*nz
						std::size_t nx = std::max((std::size_t)1, (std::size_t) pow(num_cells/6, 1.0/3.0));
						std::size_t ny = std::max((std::size_t)1, (std::size_t) sqrt(num_cells/(6*nx)));
						std::size_t nz = std::max((std::size_t)1, num_cells/(6*nx*ny));
						equivalent_mesh.reset(new dolfin::UnitCubeMesh(nx,ny,nz));
					}
					else {
						dolfin::warning(
							"could not determine mesh type of test problem (probably loaded from file), "
							"therefore direct initialization time of fine mesh could not be determined and will be set to "
							"time needed for mesh adaption");
						init_fine_mesh_skipped = true;
						init_fine_mesh_time = FMGTimer::get_max("adapt_mesh");
						nint = 1;
						break;
					}

					init_fine_mesh_time = dolfin::time()-t0;
					nint ++;
				}
				while(init_fine_mesh_time < t_int);

				init_fine_mesh_time /= nint;

				if (!init_fine_mesh_skipped) {
					// check that number of cells match
					if (equivalent_mesh->num_cells() != num_cells) {
						dolfin::warning(
							"could not reproduce fine mesh with same number of cells as adapted mesh, "
							"some initialization timings may be incorrect by the oder of %g%%",
							100.0*(equivalent_mesh->num_cells()-num_cells)/(double) num_cells);
					}
				}
			}

			// compute total losses

			t_init_mg = FMGTimer::get_sum("adapt_problem") + FMGTimer::get_min("adapt_problem");
			t_direct[i] = init_fine_mesh_time +
				FMGTimer::get_max("adapt_problem") +
				FMGTimer::get_max("init_boundary_conditions") +
				FMGTimer::get_max("init_system_lhs") +
				FMGTimer::get_max("init_system_rhs");
			t_adapt_mesh = std::max(init_fine_mesh_time, FMGTimer::get_sum("adapt_mesh"));
			t_mg[i] = t_adapt_mesh + t_init_mg + FMGTimer::get_sum("update_level");
			t_loss[i] = t_mg[i]-t_direct[i];
			t_loss_mesh = (t_adapt_mesh - init_fine_mesh_time);
			t_loss_init = (t_init_mg - FMGTimer::get_max("adapt_problem"));
			t_loss_asm_bc = FMGTimer::get_sum("init_boundary_conditions") - FMGTimer::get_max("init_boundary_conditions");
			t_loss_asm_lhs = FMGTimer::get_sum("init_system_lhs") - FMGTimer::get_max("init_system_lhs");
			t_loss_asm_rhs = FMGTimer::get_sum("init_system_rhs") - FMGTimer::get_max("init_system_rhs");
			t_loss_pro = FMGTimer::get_sum("init_prolongation");
			t_loss_sr = FMGTimer::get_sum("init_reduced_system");
			t_loss_other = t_loss[i]-(t_loss_mesh+t_loss_init+t_loss_asm_bc+t_loss_asm_lhs+t_loss_asm_rhs+t_loss_pro+t_loss_sr);

			/*
			std::cout << "update_level: " << FMGTimer::get_sum("update_level") << " " 
				<< "init_smoothers: " << FMGTimer::get_sum("init_smoothers") << " " 
				<< "init_system_lhs: " << FMGTimer::get_sum("init_system_lhs") << " " 
				<< "init_system_rhs: " << FMGTimer::get_sum("init_system_rhs") << " " 
				<< "init_boundary_conditions: " << FMGTimer::get_sum("init_boundary_conditions") << " " 
				<< "init_user: " << FMGTimer::get_sum("init_user") << " " 
				<< "init_prolongation: " << FMGTimer::get_sum("init_prolongation") << " " 
				<< "init_reduced_system: " << FMGTimer::get_sum("init_reduced_system") << " " 
				<< "init_rhs: " << FMGTimer::get_sum("init_rhs") << " " 
				<< "init_vectors: " << FMGTimer::get_sum("init_vectors") << " " 
				<< "init_fine_mesh_time: " << init_fine_mesh_time << " " 
				<< "t_direct[i]: " << t_direct[i] << std::endl;
			*/
		}

		// output details as table for last refinement

		if ((bool) parameters["show_help"]) {
			std::cout << "\n\nThe following table computes the additional time that is needed for the fmg solver "
					"with " << nrefine << " refinements to initialize in contrast to using a conventional solver "
					"(with no additional initialization time) on the fine mesh directly.\n";
		}

		Table tab(table_format);

		tab.add_text_column("name", "name", 15);
		tab.add_text_column("description", "description", 61);
		tab.add_float_column("time", "time");
		tab.add_float_column("relative", "loss %");
		tab.finish_header();

		// loss by mesh creation

		tab["name"] << "t_mesh";
		tab["description"] << "direct problem mesh creation";
		tab["time"] << init_fine_mesh_time;
		tab.finish_row();

		tab["name"] << "t_mesh_mg";
		tab["description"] << "multigrid adaptive mesh creation";
		tab["time"] << t_adapt_mesh;
	//	tab["relative"] << (FMGTimer::get_sum("adapt_mesh")/init_fine_mesh_time-1.0)*100.0;
		tab.finish_row();

		tab["name"] << "t_loss_mesh";
		tab["description"] << "loss by mesh creation";
		tab["time"] << t_loss_mesh;
		tab["relative"] << (t_loss_mesh/t_loss[nrefine])*100.0;
		tab.finish_row();

		tab.add_delimiter();

		// loss by problem initialization

		tab["name"] << "t_init";
		tab["description"] << "direct problem initialization (without t_mesh)";
		tab["time"] << FMGTimer::get_max("adapt_problem");
		tab.finish_row();

		tab["name"] << "t_init_mg";
		tab["description"] << "multigrid adaptive problem initialization (without t_mesh_mg)";
		tab["time"] << t_init_mg; // adapt + approximative coarse level init
	//	tab["relative"] << (FMGTimer::get_sum("adapt_problem")/FMGTimer::get_max("adapt_problem")-1.0)*100.0;
		tab.finish_row();

		tab["name"] << "t_loss_init";
		tab["description"] << "loss by problem initialization";
		tab["time"] << t_loss_init;
		tab["relative"] << (t_loss_init/t_loss[nrefine])*100.0;
		tab.finish_row();

		tab.add_delimiter();

		// loss by initialization of boundary values

		tab["name"] << "t_asm_bc";
		tab["description"] << "direct problem initialization of boundary values";
		tab["time"] << FMGTimer::get_max("init_boundary_conditions");
		tab.finish_row();

		tab["name"] << "t_asm_bc_mg";
		tab["description"] << "multigrid problem initialization of boundary values";
		tab["time"] << FMGTimer::get_sum("init_boundary_conditions");
	//	tab["relative"] << (FMGTimer::get_sum("init_boundary_conditions")/FMGTimer::get_max("init_boundary_conditions")-1.0)*100.0;
		tab.finish_row();

		tab["name"] << "t_loss_asm_bc";
		tab["description"] << "loss by initialization of boundary values";
		tab["time"] << t_loss_asm_bc;
		tab["relative"] << (t_loss_asm_bc/t_loss[nrefine])*100.0;
		tab.finish_row();

		tab.add_delimiter();

		// loss by assembly of lhs

		tab["name"] << "t_asm_lhs";
		tab["description"] << "direct problem assembly of lhs";
		tab["time"] << FMGTimer::get_max("init_system_lhs");
		tab.finish_row();

		tab["name"] << "t_asm_lhs_mg";
		tab["description"] << "multigrid problem assembly of lhs";
		tab["time"] << FMGTimer::get_sum("init_system_lhs");
	//	tab["relative"] << (FMGTimer::get_sum("init_system_lhs")/FMGTimer::get_max("init_system_lhs")-1.0)*100.0;
		tab.finish_row();

		tab["name"] << "t_loss_asm_lhs";
		tab["description"] << "loss by assembly of lhs";
		tab["time"] << t_loss_asm_lhs;
		tab["relative"] << (t_loss_asm_lhs/t_loss[nrefine])*100.0;
		tab.finish_row();

		tab.add_delimiter();

		// loss by assembly of rhs

		tab["name"] << "t_asm_rhs";
		tab["description"] << "direct problem assembly of rhs";
		tab["time"] << FMGTimer::get_max("init_system_rhs");
		tab.finish_row();

		tab["name"] << "t_asm_rhs_mg";
		tab["description"] << "multigrid problem assembly of rhs";
		tab["time"] << FMGTimer::get_sum("init_system_rhs");
	//	tab["relative"] << (FMGTimer::get_sum("init_system_rhs")/FMGTimer::get_max("init_system_rhs")-1.0)*100.0;
		tab.finish_row();

		tab["name"] << "t_loss_asm_rhs";
		tab["description"] << "loss by assembly of rhs";
		tab["time"] << t_loss_asm_rhs;
		tab["relative"] << (t_loss_asm_rhs/t_loss[nrefine])*100.0;
		tab.finish_row();

		tab.add_delimiter();

		// loss by prolongation operator creation

		tab["name"] << "t_loss_pro";
		tab["description"] << "loss by prolongation operator creation";
		tab["time"] << t_loss_pro;
		tab["relative"] << (t_loss_pro/t_loss[nrefine])*100.0;
		tab.finish_row();

		// loss by system reduction

		tab["name"] << "t_loss_sr";
		tab["description"] << "loss by system reduction";
		tab["time"] << t_loss_sr;
		tab["relative"] << (t_loss_sr/t_loss[nrefine])*100.0;
		tab.finish_row();

		// other losses

//		double t_loss_other = FMGTimer::get_sum("init_rhs") + FMGTimer::get_sum("init_vectors") + FMGTimer::get_sum("init_smoothers");
		tab["name"] << "t_loss_other";
		tab["description"] << "other losses";
		tab["time"] << t_loss_other;
		tab["relative"] << (t_loss_other/t_loss[nrefine])*100.0;
		tab.finish_row();

		tab.add_delimiter();

		// total

		tab["name"] << "t_direct";
		tab["description"] << "total direct problem initialization (\"ready to solve\")";
		tab["time"] << t_direct[nrefine];
		tab.finish_row();

		tab["name"] << "t_mg";
		tab["description"] << "total multigrid problem initialization (\"ready to solve\")";
		tab["time"] << t_mg[nrefine];
	//	tab["relative"] << (t_mg[i]/t_direct[i]-1.0)*100.0;
		tab.finish_row();

		// total losses

		tab.add_delimiter();

		tab["name"] << "t_loss";
		tab["description"] << "total losses";
		tab["time"] << t_loss[nrefine];
	//	tab["relative"] << (t_loss[nrefine]/t_direct[i])*100.0;
		tab["relative"] << 100.0;
		tab.finish_row();

		// list t_loss for each refinement level
		{
			if ((bool) parameters["show_help"]) {
				std::cout << "\n\nThe following table shows t_loss for each refinement level.\n";
			}

			Table tab(table_format);

			tab.add_int_column("nrefine", "refinements", 2);
			tab.add_int_column("dofs", "dofs", 7);
			tab.add_float_column("t_direct", "t_direct");
			tab.add_float_column("t_mg", "t_mg");
			tab.add_float_column("t_loss", "t_loss");
			tab.add_float_column("t_lossp", "t_loss %");
			tab.finish_header();

			// loss by mesh creation

			for (std::size_t i = 0; i <= nrefine; i++)
			{
				std::size_t dofs = mg_solver.problem().level(i).A->size(0);
				tab["nrefine"] << i;
				tab["dofs"] << dofs;
				tab["t_direct"] << t_direct[i];
				tab["t_mg"] << t_mg[i];
				tab["t_loss"] << t_loss[i];
				tab["t_lossp"] << t_loss[i]/t_direct[i]*100.0;
				tab.finish_row();
			}
		}
	}

#if 0
	Utils::print_system("A0_0", *_problem->level(0).A0, *_problem->level(0).b0);
	Utils::print_system("A_0", *_problem->level(0).A, *_problem->level(0).bp);
	//Utils::print_system("A0_1", *_problem->level(1).A0, *_problem->level(1).b0);
	//Utils::print_system("A_1", *_problem->level(1).A, *_problem->level(1).bp);
	exit(0);
#endif

	// test: residuals for cg+fmg solver
	if (1)
	{
		int nres = (int) parameters["residual_count"];
		bool ok = (nres != 0);

		try {
			boost::shared_ptr<dolfin::GenericVector> x = mg_solver.create_reduced_vector();
			boost::shared_ptr<dolfin::GenericLinearSolver> solver = create_solver(mg_solver, parameters["primary_solver"], "fmg");

			if (nres >= 0) {
				solver->parameters["relative_tolerance"] = 0.0;
				solver->parameters["absolute_tolerance"] = 0.0;
				solver->parameters["error_on_nonconvergence"] = false;
				solver->parameters["maximum_iterations"] = nres;
			}

			mg_solver.log_residuals();
			solver->solve(*x, *(mg_solver.problem_rhs()));
		}
		catch (...) {
			ok = false;
		}

		if (ok && (mg_solver._nat_residuals.size() > 0))
		{
			if ((bool) parameters["show_help"]) {
				std::cout << "\n\nThe following table shows the residuals during the " + parameters["primary_solver"].value_str() + "-iteration in different norms "
						"for the " + parameters["primary_solver"].value_str() + "+fmg solver on the finest level. " <<
						"The second table shows the relative residuals. " <<
						"The third table shows the logarithmic (base 10) reduction factors of the residual for each iteration (e.g. -1.0 means reduction by one decimal power).\n";
			}

			typedef std::map<std::string, std::vector<double> > norm_map;
			norm_map norms;
			norms.insert(std::make_pair("l2", mg_solver._l2_residuals));
			norms.insert(std::make_pair("linf", mg_solver._linf_residuals));
			norms.insert(std::make_pair("prec_l2", mg_solver._prec_l2_residuals));
			norms.insert(std::make_pair("prec_linf", mg_solver._prec_linf_residuals));
			norms.insert(std::make_pair("natural", mg_solver._nat_residuals));

			{
				Table tab(table_format);

				tab.add_int_column("iter", "iter", 5);
				for (norm_map::iterator i = norms.begin(); i != norms.end(); i++) {
					tab.add_float_column(i->first, i->first);
				}
				tab.finish_header();

				for (std::size_t k = 0; k < mg_solver._nat_residuals.size(); k++)
				{
					tab["iter"] << k;
					for (norm_map::iterator i = norms.begin(); i != norms.end(); i++) {
						tab[i->first] << i->second[k];
					}
					tab.finish_row();
				}
			}
			{
				Table tab(table_format);

				tab.add_int_column("iter", "iter", 5);
				for (norm_map::iterator i = norms.begin(); i != norms.end(); i++) {
					tab.add_float_column(i->first, i->first);
				}
				tab.finish_header();

				for (std::size_t k = 0; k < mg_solver._nat_residuals.size(); k++)
				{
					tab["iter"] << k;
					for (norm_map::iterator i = norms.begin(); i != norms.end(); i++) {
						tab[i->first] << i->second[k]/i->second[0];
					}
					tab.finish_row();
				}
			}
			{
				Table tab(table_format);

				tab.add_int_column("iter", "iter", 5);
				for (norm_map::iterator i = norms.begin(); i != norms.end(); i++) {
					tab.add_float_column(i->first, i->first);
				}
				tab.finish_header();

				for (std::size_t k = 1; k < mg_solver._nat_residuals.size(); k++)
				{
					tab["iter"] << k;
					for (norm_map::iterator i = norms.begin(); i != norms.end(); i++) {
						tab[i->first] << log10(i->second[k]/i->second[k-1]);
					}
					tab.finish_row();
				}
			}
		}

		mg_solver.log_residuals(false);
		mg_solver.parameters = fmg_params;
	}
	// test: initialization time vs number of refinements for all preconditioned methods listed in test_solver parameter
	if (1)
	{
		// get list of solver+preconditioner combinations
		std::vector<std::pair<std::string, std::string> > solver_prec_pairs;
		std::vector<std::string> solver_prec;
		std::vector<bool> is_ks;
		std::string test_solver = parameters["test_solver"];
		boost::replace_all(test_solver, "*", parameters["primary_solver"].value_str());
		Utils::split_string(solver_prec, test_solver);
		solver_prec.push_back(parameters["primary_solver"].value_str() + "+fmg");

		// remove duplicate solver
		solver_prec.resize(std::unique(solver_prec.begin(), solver_prec.end()) - solver_prec.begin());

		solver_prec_pairs.resize(solver_prec.size());
		is_ks.resize(solver_prec.size(), true);
		std::size_t k = 0;
		for (std::size_t i = 0; i < solver_prec.size(); i++) {
			std::vector<std::string> parts;
			Utils::split_string(parts, solver_prec[i], "+");
			solver_prec_pairs[k].first = parts[0];
			solver_prec_pairs[k].second = (parts.size()>1) ? parts[1] : "none";
			try {
				// skip all direct and fmg solver
			//	if (solver_prec_pairs[i].first == "fmg") continue;
			//	if (solver_prec_pairs[i].second == "fmg") continue;
			//	if (solver_prec_pairs[i].second == "none") continue;
				boost::shared_ptr<dolfin::GenericLinearSolver> s = create_solver(mg_solver, solver_prec_pairs[i].first, solver_prec_pairs[i].second);
				if (dynamic_cast<dolfin::PETScLUSolver*>(s.get()) != NULL) continue;
			}
			catch (...) { continue; }
			k++;
		}
		solver_prec_pairs.resize(k);
		
		if (k > 0)
		{
			if ((bool) parameters["show_help"]) {
				std::cout << "\n\nThe following table shows the solver initialization time per dof and solve time per dof " <<
				"in microseconds (1e-6).\n";
			}

			mg_solver.parameters["coarse_levels_rhs_assembly"] = true;
			mg_solver.update(fmg::MultigridSolver::UpdateFlags::system_rhs);
			const std::vector<boost::shared_ptr<MultigridLevel> >& levels = mg_solver.problem().levels();

			Table tab(table_format);

			tab.add_int_column("nrefine", "refinements", 2);
			tab.add_int_column("dofs", "dofs", 7);
			for (std::size_t i = 0; i < solver_prec.size(); i++) {
				tab.add_float_column(solver_prec[i] + "-init", solver_prec[i] + " init");
				tab.add_float_column(solver_prec[i] + "-time", solver_prec[i] + " time");
			}
			tab.finish_header();

			for (std::size_t j = 0; j <= nrefine; j++)
			{
				int level_offset = (int)(nrefine-j);
				mg_solver.parameters["level_offset"] = level_offset;

				boost::shared_ptr<dolfin::GenericVector> x = mg_solver.create_reduced_vector();
				tab["nrefine"] << j;
				tab["dofs"] << x->size();

				for (std::size_t i = 0; i < solver_prec.size(); i++)
				{
					mg_solver.parameters = fmg_params;
					mg_solver.parameters["level_offset"] = level_offset;

					try {
						double t_init[2];
						int n[2];
						std::size_t niter = 0;
						double t0 = dolfin::time();
						//int m = 1;
						
						t_init[0] = t_init[1] = 0;
						
						do {
							for (int k = 0; k < 2; k++)
							{
								x->zero();
								for (std::size_t l = 0; l < levels.size(); l++) {
									levels[l]->coarse_solver.reset();
								}
								
								double t1 = dolfin::time();
								boost::shared_ptr<dolfin::GenericLinearSolver> solver = create_solver(mg_solver, 
									solver_prec_pairs[i].first, solver_prec_pairs[i].second);
								
								if (k == 0) {
									solver->parameters["relative_tolerance"] = (double) fmg_params["relative_tolerance"];
									solver->parameters["absolute_tolerance"] = (double) fmg_params["absolute_tolerance"];
									solver->parameters["error_on_nonconvergence"] = true;
									solver->parameters["maximum_iterations"] = (int) fmg_params["maximum_iterations"];
								}
								else {
									solver->parameters["relative_tolerance"] = 0.0;
									solver->parameters["absolute_tolerance"] = 0.0;
									solver->parameters["error_on_nonconvergence"] = false;
									solver->parameters["maximum_iterations"] = (int)(n[0]>1 ? (n[0]/2) : (n[0]+2));
								}
								
								n[k] = (int) solver->solve(*x, *(mg_solver.problem_rhs()));

								t_init[k] += dolfin::time()-t1;
							}
							niter++;
						}
						while ((dolfin::time()-t0) < t_int);
						
						t_init[0] /= niter;
						t_init[1] /= niter;

						// t_init[0] = (n[0]+1)*ts + ti
						// t_init[1] = (n[1]+1)*ts + ti
						//std::cout << solver_prec[i] << ": " << t_init[0] << " " << t_init[1] << " " << n[0] << " "<< n[1] << " " << std::endl;

						// it is currently difficult to tell how much times the preconditioner will be
						// applied, in cg this is maximum_iterations+1,
						double ts, ti;
						
						if (n[1] == n[0]) {
							ts = (t_init[1])/(double)(n[1]+1);
							ti = 0.0;
						}
						else {
							ts = (t_init[1]-t_init[0])/(double)(n[1]-n[0]);
							ti = (t_init[1]+t_init[0]-(n[0]+n[1]+2)*ts)*0.5;
						}
						
						// for multigrid methods we add the t_loss to the initialization time.
						if (solver_prec_pairs[i].first == "fmg" || solver_prec_pairs[i].second == "fmg") {
							ti += t_loss[j];
						}
						
						tab[solver_prec[i] + "-init"] << (1e6*ti/x->size());
						tab[solver_prec[i] + "-time"] << (1e6*ts*(n[0]+1)/x->size());
					}
					catch (...) {
						tab[solver_prec[i] + "-init"] << "-";
						tab[solver_prec[i] + "-time"] << "-";
					}
				}

				tab.finish_row();
			}

			mg_solver.parameters = fmg_params;
		}
	}

	// test: number of iterations and solve time vs number of refinements for all methods listed in test_solver parameter
	if (1)
	{
		if ((bool) parameters["show_help"]) {
			std::cout << "\n\nThe following table shows the number of iterations and solve time versus the number of mesh refinements "
					"for all methods listed in the \"test_solver\" parameter. "
					"All solver timings don't include the time for initialization of the solver. You have to add the t_loss from the "
					"previous table to the timings for all solvers using fmg to have the total time that is required when doing "
					"only a single solve.\n";
		}

		mg_solver.parameters["coarse_levels_rhs_assembly"] = true;
		mg_solver.update(fmg::MultigridSolver::UpdateFlags::system_rhs);

		// get list of solver+preconditioner combinations
		std::vector<std::pair<std::string, std::string> > solver_prec_pairs;
		std::vector<std::string> solver_prec;
		std::vector<bool> is_ks;
		std::string test_solver = parameters["test_solver"];
		boost::replace_all(test_solver, "*", parameters["primary_solver"].value_str());
		Utils::split_string(solver_prec, test_solver);
		solver_prec.push_back(parameters["primary_solver"].value_str() + "+fmg");

		// remove duplicate solver
		solver_prec.resize(std::unique(solver_prec.begin(), solver_prec.end()) - solver_prec.begin());

		solver_prec_pairs.resize(solver_prec.size());
		is_ks.resize(solver_prec.size(), true);
		for (std::size_t i = 0; i < solver_prec.size(); i++) {
			std::vector<std::string> parts;
			Utils::split_string(parts, solver_prec[i], "+");
			solver_prec_pairs[i].first = parts[0];
			solver_prec_pairs[i].second = (parts.size()>1) ? parts[1] : "none";
			try {
				boost::shared_ptr<dolfin::GenericLinearSolver> s = create_solver(mg_solver, solver_prec_pairs[i].first, solver_prec_pairs[i].second);
				is_ks[i] = (dynamic_cast<dolfin::PETScLUSolver*>(s.get()) == NULL);
			}
			catch (...) { }
			test_fmg = test_fmg || (solver_prec_pairs[i].first == "fmg");
		}

		Table tab(table_format);

		tab.add_int_column("nrefine", "refinements", 2);
		tab.add_int_column("dofs", "dofs", 7);
		for (std::size_t i = 0; i < solver_prec.size(); i++) {
			if (is_ks[i]) tab.add_int_column(solver_prec[i] + "-iter", solver_prec[i] + " iter");
			tab.add_float_column(solver_prec[i] + "-time", solver_prec[i] + " time");
		}
		tab.finish_header();

		for (std::size_t j = 0; j <= nrefine; j++)
		{
			int level_offset = (int)(nrefine-j);
			mg_solver.parameters["level_offset"] = level_offset;

			boost::shared_ptr<dolfin::GenericVector> x = mg_solver.create_reduced_vector();
			tab["nrefine"] << j;
			tab["dofs"] << x->size();

			bool vtkSaved = false;

			for (std::size_t i = 0; i < solver_prec.size(); i++)
			{
				mg_solver.parameters = fmg_params;
				mg_solver.parameters["level_offset"] = level_offset;

				try {
					double t_solve;
					std::size_t n = integrated_solve(mg_solver, solver_prec_pairs[i].first, solver_prec_pairs[i].second,
							*x, *(mg_solver.problem_rhs()), t_int, t_solve);
					if (is_ks[i]) tab[solver_prec[i] + "-iter"] << n;
					tab[solver_prec[i] + "-time"] << t_solve;

					try {
						if (!vtkSaved && (bool)parameters["write_solutions"]) {
							std::string format = parameters["solution_file_format"].value_str();
							std::string filename = (boost::format(format) % j).str();
							dolfin::File file(filename);
							mg_solver.expand_vector(*x, *(mg_solver.solution()->vector()));
							file << *(mg_solver.solution());
							vtkSaved = true;
						}
					}
					catch (...) {
					}
				}
				catch (...) {
					if (is_ks[i]) tab[solver_prec[i] + "-iter"] << "-";
					tab[solver_prec[i] + "-time"] << "-";
				}
			}

			tab.finish_row();
		}

		mg_solver.parameters = fmg_params;
	}

	// test: number of iterations and solve time vs smoothing method/number of iterations for cg+fmg solver
	if (1)
	{
		// get list of pre-smmother+post-smmother combinations
		std::vector<std::pair<std::string, std::string> > smoother_pairs;
		std::vector<std::pair<double, double> > relax_pairs;
		std::vector<std::string> smoother;
		Utils::split_string(smoother, parameters["test_smoothers"].value_str());
		std::size_t max_smoother_width = 0;

		// remove duplicate smoothers
		smoother.resize(std::unique(smoother.begin(), smoother.end()) - smoother.begin());

		if (smoother.size() > 0)
		{
			if ((bool) parameters["show_help"]) {
				std::cout << "\n\nThe following table shows the number of iterations and solve time for different smoother "
						"methods (as listed in the \"test_smoothers\" parameter up to \"max_smoothing_reps\" repetitions) "
						"for the " + parameters["primary_solver"].value_str() + "+fmg solver on the finest level.\n";
			}

			smoother_pairs.resize(smoother.size());
			relax_pairs.resize(smoother.size());
			for (std::size_t i = 0; i < smoother.size(); i++)
			{
				max_smoother_width = std::max(max_smoother_width, smoother[i].size());

				std::vector<std::string> parts;
				std::vector<std::string> parts2;
				Utils::split_string(parts, smoother[i], "+");

				Utils::split_string(parts2, parts[0], "@");
				smoother_pairs[i].first = parts2[0];
				relax_pairs[i].first = (parts2.size()>1) ? atof(parts2[1].c_str()) : 1.0;

				if (parts.size()>1) {
					Utils::split_string(parts2, parts[1], "@");
					smoother_pairs[i].second = parts2[0];
					relax_pairs[i].second = (parts2.size()>1) ? atof(parts2[1].c_str()) : 1.0;
				}
				else {
					smoother_pairs[i].second = smoother_pairs[i].first;
					relax_pairs[i].second = relax_pairs[i].first;
				}
			}

			Table tab(table_format);
			TestPattern tp(tab, mg_solver, parameters, test_fmg);

			tab.add_text_column("smoother", "smoother", max_smoother_width);
			tab.add_int_column("reps", "reps");
			tp.add_columns();
			tab.finish_header();

			int max_smoothing_reps = parameters["max_smoothing_reps"];

			for (std::size_t i = 0; i < smoother.size(); i++)
			{
				//std:: cout << smoother_pairs[i].first << " " << smoother_pairs[i].second << " " <<  relax_pairs[i].first << " " <<  relax_pairs[i].second << std::endl;
				mg_solver.parameters["pre_smoother"] = smoother_pairs[i].first;
				mg_solver.parameters["post_smoother"] = smoother_pairs[i].second;
				mg_solver.parameters["pre_smoother_relax"] = relax_pairs[i].first;
				mg_solver.parameters["post_smoother_relax"] = relax_pairs[i].second;
				mg_solver.update(MultigridSolver::UpdateFlags::smoothers);

				if (i > 0) tab.add_delimiter();

				for (int j = 1; j <= max_smoothing_reps; j++)
				{
					mg_solver.parameters["pre_smoother_reps"] = j;
					mg_solver.parameters["post_smoother_reps"] = j;

					tab["smoother"] << smoother[i];
					tab["reps"] << j;
					tp.run();
					tab.finish_row();
				}
			}

			mg_solver.parameters = fmg_params;
			mg_solver.update(MultigridSolver::UpdateFlags::smoothers);
		}
	}

	// test: number of iterations and solve time vs coarse solver for cg+fmg solver
	if (1)
	{
		// get list of solver+preconditioner combinations
		std::vector<std::pair<std::string, std::string> > solver_prec_pairs;
		std::vector<std::string> solver_prec;
		std::string test_coarse_solver = parameters["test_coarse_solver"];
		boost::replace_all(test_coarse_solver, "*", parameters["primary_solver"].value_str().c_str());
		Utils::split_string(solver_prec, test_coarse_solver);
		std::size_t max_solver_prec_width = 0;

		// remove duplicate solvers
		solver_prec.resize(std::unique(solver_prec.begin(), solver_prec.end()) - solver_prec.begin());

		if (solver_prec.size() > 0)
		{
			if ((bool) parameters["show_help"]) {
				std::cout << "\n\nThe following table shows the number of iterations and solve time for different solvers for "
						"the coarse level (as listed in the \"test_coarse_solver\" parameter) "
						"for the " + parameters["primary_solver"].value_str() + "+fmg solver on the finest level.\n";
			}

			solver_prec_pairs.resize(solver_prec.size());
			for (std::size_t i = 0; i < solver_prec.size(); i++) {
				std::vector<std::string> parts;
				Utils::split_string(parts, solver_prec[i], "+");
				solver_prec_pairs[i].first = parts[0];
				solver_prec_pairs[i].second = (parts.size()>1) ? parts[1] : "none";
				max_solver_prec_width = std::max(max_solver_prec_width, solver_prec[i].size());
			}

			Table tab(table_format);
			TestPattern tp(tab, mg_solver, parameters, test_fmg);

			tab.add_text_column("solver", "coarse solver", max_solver_prec_width);
			tp.add_columns();
			tab.finish_header();

			std::vector<std::pair<std::string, std::string> > lus_methods = dolfin::PETScLUSolver::methods();

			for (std::size_t i = 0; i < solver_prec.size(); i++)
			{
				try {
					if (std::find_if(lus_methods.begin(), lus_methods.end(), FindFirst(solver_prec_pairs[i].first)) != lus_methods.end()) {
						mg_solver.parameters["coarse_solver_type"] = "lu";
						mg_solver.parameters["lu_solver_method"] = solver_prec_pairs[i].first;
					}
					else {
						mg_solver.parameters["coarse_solver_type"] = "krylov";
						mg_solver.parameters["krylov_solver_method"] = solver_prec_pairs[i].first;
						mg_solver.parameters["krylov_solver_preconditioner"] = solver_prec_pairs[i].second;
					}
				}
				catch (...) {
					continue;
				}

				tab["solver"] << solver_prec[i];
				tp.run();
				tab.finish_row();
			}

			mg_solver.parameters = fmg_params;
		}
	}

	// test: number of iterations and solve time vs cycle pattern for cg+fmg solver
	if (1)
	{
		// get list of pre-smmother+post-smmother combinations
		std::vector<std::string> patterns;
		Utils::split_string(patterns, parameters["test_cycle_patterns"].value_str());
		std::size_t max_pattern_width = 0;

		// remove duplicate patterns
		patterns.resize(std::unique(patterns.begin(), patterns.end()) - patterns.begin());

		if (patterns.size() > 0)
		{
			if ((bool) parameters["show_help"]) {
				std::cout << "\n\nThe following table shows the number of iterations and solve time for different cycle patterns "
						"(as listed in the \"test_cycle_patterns\" parameter) "
						"for the " + parameters["primary_solver"].value_str() + "+fmg solver on the finest level.\n";
			}

			for (std::size_t i = 0; i < patterns.size(); i++) {
				max_pattern_width = std::max(max_pattern_width, patterns[i].size());
			}

			Table tab(table_format);
			TestPattern tp(tab, mg_solver, parameters, test_fmg);

			tab.add_text_column("pattern", "pattern", max_pattern_width);
			tp.add_columns();
			tab.finish_header();

			for (std::size_t i = 0; i < patterns.size(); i++)
			{
				try {
					// will throw if pattern not suitable for number of refinements
					fmg::VectorCyclePattern pattern(patterns[i], nrefine+1);
				}
				catch (...) {
					continue;
				}

				mg_solver.parameters["cycle_pattern"] = patterns[i];

				tab["pattern"] << patterns[i];
				tp.run();
				tab.finish_row();
			}

			mg_solver.parameters = fmg_params;
		}
	}

	// test: number of iterations and solve time vs cycle depth

	std::size_t min_coarse_depth = parameters["min_coarse_depth"];

	if (nrefine >= min_coarse_depth)
	{
		// generate list of patterns
		std::vector<std::string> patterns;
		std::string pattern;

		pattern = "1";
		for (std::size_t i = 0; i <= nrefine; i++) {
			if (i >= min_coarse_depth) patterns.push_back(pattern);
			pattern += "/1";
		}

		if (patterns.size() > 0)
		{
			if ((bool) parameters["show_help"]) {
				std::cout << "\n\nThe following table shows the number of iterations and solve time for different cycle depths "
						"(the number of coarsenings relative to the finest level) "
						"for the " + parameters["primary_solver"].value_str() + "+fmg solver on the finest level.\n";
			}

			std::size_t max_pattern_width = 0;

			for (std::size_t i = 0; i < patterns.size(); i++) {
				max_pattern_width = std::max(max_pattern_width, patterns[i].size());
			}

			Table tab(table_format);
			TestPattern tp(tab, mg_solver, parameters, test_fmg);

			tab.add_int_column("depth", "depth");
			tab.add_text_column("pattern", "pattern", max_pattern_width);
			tab.add_int_column("cdofs", "coarse dofs");
			tp.add_columns();
			tab.finish_header();

			for (std::size_t i = 0; i < patterns.size(); i++)
			{
				try {
					// will throw if pattern not suitable for number of refinements
					fmg::VectorCyclePattern pattern(patterns[i], nrefine+1);
				}
				catch (...) {
					continue;
				}

				std::size_t depth = (patterns[i].size()-1)/2;

				mg_solver.parameters["cycle_pattern"] = patterns[i];

				tab["depth"] << depth;
				tab["pattern"] << patterns[i];
				tab["cdofs"] << mg_solver.problem().level(nrefine - depth).A->size(0);
				tp.run();
				tab.finish_row();
			}

			mg_solver.parameters = fmg_params;
		}
	}

	// test: solve time for different operations of the multigrid algorithm
	if (1)
	{
		mg_solver.parameters["tolerance_checking"] = false;
		mg_solver.parameters["error_on_nonconvergence"] = false;

		boost::shared_ptr<dolfin::GenericVector> x = mg_solver.create_reduced_vector();
		bool failed = false;
		double t_solve = 0;
		std::size_t niter = 0;
		std::size_t s = 0;

		FMGTimer::reset_all();

		try {
			double t_solve0 = dolfin::time();
			do {
				mg_solver.parameters["maximum_iterations"] = 1 << s;
				x->zero();
				niter += mg_solver.solve(*x, false);
				s++;
				t_solve = dolfin::time() - t_solve0;
			}
			while (t_solve < t_int);
		}
		catch (...) {
			failed = true;
		}

		if (!failed)
		{
			std::vector<std::pair<std::string,std::string> > tasks = boost::assign::pair_list_of
				("calc_defect_level", "calc defect")
				("pre_smooth_level", "pre smooth")
				("post_smooth_level", "post smooth")
				("restrict_to_coarser_level", "restrict")
				("prolongate_to_finer_level", "prolongate")
				("correct_error_level", "correct error")
				("coarse_solve_level", "coarse solve")
				("test_residual_break_condition", "test break")
				;

			double sum_total = 0;
			double mean_total = 0;
			for (std::size_t i = 0; i < tasks.size(); i ++) {
				sum_total += FMGTimer::get_sum(tasks[i].first);
				mean_total += FMGTimer::get_mean(tasks[i].first);
			}

			if ((bool) parameters["show_help"]) {
				std::cout << "\n\nThe following table shows the execution time of different operations for " <<
						niter << " cycles (" << std::setprecision(2) << sum_total << "s) of the multigrid algorithm "
						"on the finest level using nonzero initial guess.\n";
			}

			Table tab(table_format);

			tab.add_text_column("op", "operation", 14);
			tab.add_int_column("reps", "reps", 7);
			tab.add_float_column("min", "min time");
			tab.add_float_column("max", "max time");
			tab.add_float_column("total", "total");
			tab.add_float_column("mean", "time/rep");
			tab.add_float_column("meanp", "time/rep %");
			tab.add_float_column("itime", "time/iter");
			tab.add_float_column("itimep", "time/iter %");
			tab.finish_header();

			for (std::size_t i = 0; i < tasks.size(); i ++) {
				tab["op"] << tasks[i].second;
				tab["reps"] << FMGTimer::get_reps(tasks[i].first);
				tab["min"] << FMGTimer::get_min(tasks[i].first);
				tab["max"] << FMGTimer::get_max(tasks[i].first);
				tab["total"] << FMGTimer::get_sum(tasks[i].first);
				tab["mean"] << FMGTimer::get_mean(tasks[i].first);
				tab["meanp"] << (FMGTimer::get_mean(tasks[i].first)/mean_total*100.0);
				tab["itime"] << (FMGTimer::get_sum(tasks[i].first)/niter);
				tab["itimep"] << (FMGTimer::get_sum(tasks[i].first)/sum_total*100.0);
				tab.finish_row();
			}
		}
		else {
			std::cout << "Test failed!\n";
		}

		mg_solver.parameters = fmg_params;
	}

	dolfin::LogManager::logger.set_log_active(log_active);
}


std::size_t Tests::integrated_solve(MultigridSolver& mg_solver, const std::string& method, const std::string& prec,
		dolfin::GenericVector& x, const dolfin::GenericVector& b, double t_int, double& t_solve)
{
	const std::vector<boost::shared_ptr<MultigridLevel> >& levels = mg_solver.problem().levels();
	std::size_t niter = 0;
	std::size_t n;

	t_solve = 0;
	
	do {
		// recreate solver and reset mg coarse solvers (maybe they hold some factorization)
		boost::shared_ptr<dolfin::GenericLinearSolver> solver = create_solver(mg_solver, method, prec);
		for (std::size_t i = 0; i < levels.size(); i++) {
			levels[i]->coarse_solver.reset();
		}

		// solve
		double t0 = dolfin::time();
		n = solver->solve(x, b);
		t_solve += dolfin::time() - t0;
		niter++;
	}
	while (t_solve < t_int);

	t_solve /= niter;

	return n;
}


boost::shared_ptr<dolfin::GenericLinearSolver> Tests::create_solver(MultigridSolver& mg_solver,
		const std::string& method, const std::string& prec)
{
	std::vector<std::pair<std::string, std::string> > lus_methods = dolfin::PETScLUSolver::methods();
	std::vector<std::pair<std::string, std::string> > ks_methods = dolfin::PETScKrylovSolver::methods();
	boost::shared_ptr<dolfin::GenericLinearSolver> solver;

	if (method == "fmg")
	{
		if (prec != "none") {
			dolfin::error("you can't use the fmg-solver with preconditioner");
		}
		solver.reset(&mg_solver, dolfin::NoDeleter());
	}
	else
	{
		if (std::find_if(ks_methods.begin(), ks_methods.end(), FindFirst(method)) != ks_methods.end())
		{
			if (prec == "fmg") {
				solver.reset(new MultigridPreconditionedKrylovSolver(mg_solver, method));
			}
			else {
				solver.reset(new dolfin::PETScKrylovSolver(method, prec));
				solver->set_operator(mg_solver.problem_operator());
			}

			if (solver->parameters.has_parameter("convergence_norm_type")) {
				solver->parameters["convergence_norm_type"] = "preconditioned";
			}
			KSPSetNormType(*(((dolfin::PETScKrylovSolver*)solver.get())->ksp()), KSP_NORM_PRECONDITIONED);
			
			solver->parameters["relative_tolerance"] = (double) mg_solver.parameters["relative_tolerance"];
			solver->parameters["absolute_tolerance"] = (double) mg_solver.parameters["absolute_tolerance"];
		}
		else if (std::find_if(lus_methods.begin(), lus_methods.end(), FindFirst(method)) != lus_methods.end())
		{
			if (prec != "none") {
				dolfin::error("you can't use a lu-solver with preconditioner");
			}
			solver.reset(new dolfin::PETScLUSolver(method));
			solver->set_operator(mg_solver.problem_operator());
		}
		else {
			dolfin::error("unknown solver %s", method.c_str());
		}
	}

	return solver;
}


const dolfin::Parameters& Tests::default_parameters(bool linear, bool symmetric, bool uniform)
{
	typedef std::map<int, boost::shared_ptr<dolfin::Parameters> > param_map;
	static param_map _default_param_map;
	int key = (linear ? 1 : 0) & (symmetric ? 2 : 0) & (uniform ? 4 : 0);
	if (_default_param_map.count(key) > 0) return *_default_param_map[key];

	boost::shared_ptr<dolfin::Parameters> p(new dolfin::Parameters("fmg_test"));

	std::set<std::string> allowed_table_formats;

	allowed_table_formats.insert("console");
	allowed_table_formats.insert("csv");
	allowed_table_formats.insert("tab");
	allowed_table_formats.insert("latex");
	allowed_table_formats.insert("rest");

	p->add("num_refinements", 1, 0, INT_MAX);
	p->add("primary_solver", symmetric ? "cg" : "gmres");
	p->add("test_solver", "");
	p->add("test_coarse_solver", "");
	p->add("test_smoothers", "");
	p->add("test_cycle_patterns", "");
	p->add("table_format", "console", allowed_table_formats);
	p->add("integration_time", 0.0, 0.0, DBL_MAX);
	p->add("max_smoothing_reps", 1, 0, INT_MAX);
	p->add("min_coarse_depth", 1, 0, INT_MAX);
	p->add("residual_count", -1, -1, INT_MAX);
	p->add("show_help", true);
	p->add("write_solutions", false);
	p->add("solution_file_format", "solution_%d.pvd");

	p->add(MultigridSolver::default_parameters(linear, symmetric, uniform));

	_default_param_map.insert(param_map::value_type(key, p));
	return *p;
}


} // namespace
